//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.
#include "BrashException.h"
#include <utility>

#include <memory>
#include <string>
#include <vector>
#include <sys/stat.h>
#include <cxxtls/lexical_cast.h>
#include <cxxtls/simple_regex.h>
#include <cxxtls/strtool.h>
#include <fstream>

#include <cxxtls/scriptthread.h>

namespace cxxtls
{
    namespace ScriptThread 
    {
        class ThreadContext;  // forward declare a class for use as the context parameter to the execute methods
    }
};

typedef cxxtls::ScriptThread::ThreadContext  ThreadContext;

namespace BrashTestExpression
{
   using std::string;
   using std::auto_ptr;

DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxInvalidFileOp);
DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxMissingExpression);
DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxMissingFileName);
DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxBinaryExpressionMissing2ndOperand);
DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxInvalidOperand);
DECLARE_CXXTLS_EXCEPTION(TestCommandSyntaxMissingExpectedToken);
                       
struct TestExpression
    ///
    ///  A representation of a node in the test command's syntax tree.
    ///
    ///  The execute method returns either true or false to indicate that the
    ///  expression's execution was successful.
    ///
{
    TestExpression() 
    : inverted_(false)
    {
    }

    virtual ~TestExpression() {}  // we typically use self destructing objects or
                                  // auto_ptr<TestExpression> in derived classes,
                                  // but destructors are declared virtual for safety.

    virtual bool execute(ThreadContext* context) const { return true; } 
        ///<  derived classes should override this function to 
        ///   define the needed algorithm for implementing the
        ///   derived class' behavior.

    virtual string stringValue(ThreadContext* context) const { return execute(context) ? "1" : "0"; }  

    void setInverted(bool flag) { inverted_ = flag; }

protected:
    bool inverted_;  // usually false, only set to true when the not operator is used and it is
                     // only allowed at the beginning of a term

};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct AndExpression
: public TestExpression
{
    AndExpression(TestExpression *lhs, TestExpression *rhs)
    :  lhs_(lhs)
    ,  rhs_(rhs)
    {
    }

    virtual bool execute(ThreadContext *context) const
    {
       
       bool lhs  = lhs_->execute(context);

       bool rv = lhs &&  rhs_->execute(context);


       if(inverted_)
          rv = !rv;


       return rv;
    }

    virtual ~AndExpression() {} // auto ptrs delete the members


private:
    auto_ptr<TestExpression> lhs_;
    auto_ptr<TestExpression> rhs_;
};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct OrExpression
: public TestExpression
{
    OrExpression(TestExpression *lhs, TestExpression *rhs)
    :  lhs_(lhs)
    ,  rhs_(rhs)
    {
    }

    virtual bool execute(ThreadContext *context) const
    {

       bool lhs = lhs_->execute(context);


       bool rv = lhs || rhs_->execute(context);


       if(inverted_)
          rv = ! rv;

       return rv;
    }



private:
    auto_ptr<TestExpression> lhs_;
    auto_ptr<TestExpression> rhs_;
};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct StringExpression
: public TestExpression
{
    StringExpression(string const &s)
    :  string_(s)
    {
    }

    virtual bool execute(ThreadContext *context) const
    {
       bool rv =!string_.empty();

       if(inverted_)
          rv = !rv;

       return rv;
    }

    virtual string stringValue(ThreadContext *context) const { return string_; }


private:
    string string_;
};

#ifndef STRINGIFY
#define STRINGIFY(x) #x
#endif



#define DEFINE_TEST_STRING_BINOP(name, operator) \
struct name\
: public TestExpression \
{\
   name(TestExpression *lhs, TestExpression *rhs) \
   : lhs_(lhs) \
   , rhs_(rhs)\
   {\
   }\
\
   virtual ~name() {}\
\
   virtual bool execute(ThreadContext*context) const\
   {\
       bool rv =  lhs_->stringValue(context) operator rhs_->stringValue(context);\
       if(inverted_)  rv = !rv;\
       return rv;\
   }\
\
private:\
   auto_ptr<TestExpression> lhs_;\
   auto_ptr<TestExpression> rhs_;\
}

DEFINE_TEST_STRING_BINOP(StringEqualExpression, ==);
DEFINE_TEST_STRING_BINOP(StringNotEqualExpression, !=);

struct RegexEqualExpression
: public TestExpression 
{
   RegexEqualExpression(TestExpression *lhs, TestExpression *rhs) 
   : lhs_(lhs) 
   , rhs_(rhs)
   {
   }

   virtual ~RegexEqualExpression() {}

   virtual bool execute(ThreadContext*context) const
   {
       cxxtls::SimpleRegex  tmp(cxxtls::StrTool::decodeStandardEscapeSequences(rhs_->stringValue(context)));


       bool rv =  tmp(lhs_->stringValue(context));
       if(inverted_)  rv = !rv;
       return rv;
   }

private:
   auto_ptr<TestExpression> lhs_;
   auto_ptr<TestExpression> rhs_;
};




#define DEFINE_TEST_NUMERIC_STRING_BINOP(name, operator) \
struct name\
: public TestExpression \
{\
   name(TestExpression *lhs, TestExpression *rhs) \
   : lhs_(lhs) \
   , rhs_(rhs)\
   {\
   }\
\
   virtual ~name() {}\
\
   virtual bool execute(ThreadContext*context) const\
   {\
       bool rv=false; \
\
       try\
       {\
           rv =  cxxtls::lexical_cast<int>(lhs_->stringValue(context)) operator cxxtls::lexical_cast<int>(rhs_->stringValue(context));\
       }\
       catch(...)\
       {\
           rv =  (lhs_->stringValue(context)) operator (rhs_->stringValue(context));\
       }\
       if(inverted_)  rv = !rv;\
       return rv;\
   }\
\
private:\
   auto_ptr<TestExpression> lhs_;\
   auto_ptr<TestExpression> rhs_;\
}

DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringGreaterExpression, >);
DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringLessExpression, <);
DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringGreaterEqualExpression, >=);
DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringLessEqualExpression, <=);
DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringEqualExpression, ==);
DEFINE_TEST_NUMERIC_STRING_BINOP(IntStringNotEqualExpression, !=);

enum FILEOP
{
    FILEOP_UNSPEC,         // unspecified operation, do not use
    FILEOP_EXISTS,
    FILEOP_READABLE,
    FILEOP_WRITEABLE,
    FILEOP_ISDIR,
    FILEOP_NOTEMPTY,      // exist, isn't a directory, and is not empty
    FILEOP_EMPTY,         // non-existent files are empty
    FILEOP_EXE,           // is executable 
    FILEOP_ISTERM,        // operand is a string which is the number of a file handle in the ThreadContext
                          // execute() returns true if the handle is a terminal
    FILEOP_REGULAR,       // exists and is not a directory (-e)
    FILEOP_STR_NEMPTY,    // argument is just a string. return true if string not empty
    FILEOP_STR_EMPTY,     // argument is just a string, return true if it is empty
};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
struct FileOpExpression
: public TestExpression
{
    FileOpExpression(FILEOP op, string const &s)
    :  string_(s)
    ,  op_(op)
    {
    }

    static bool fileNumberIsTerminal(ThreadContext*, string const &);  ///<  returns true if the string
                                                                       ///   is a valid number that is the
                                                                       ///   handle of an open file that is a terminal

    virtual bool execute(ThreadContext*context) const
    {
       if(op_ == FILEOP_STR_NEMPTY)
       {
          return !string_.empty();
       }

       if(op_ == FILEOP_STR_EMPTY)
       {
          return string_.empty();
       }



       struct stat fileStatus;

       bool error=false;

       if(op_ != FILEOP_ISTERM)
       {
           if(0 != stat(context->fileHandles_->mapFilename(string_).c_str(), &fileStatus))
              error = true;
       }

       switch(op_)
       {
           case FILEOP_ISTERM:
              {
                  return fileNumberIsTerminal(context,string_);
              };

           case FILEOP_EXISTS:
              {

                  bool rv = !error;
                  
                  if(inverted_)
                     rv = ! rv;
                  
                  
                  return rv;
              }

           case FILEOP_READABLE:
              {

#ifdef _MSC_VER
                  bool rv = !error && (fileStatus.st_mode & S_IREAD);
#else
                  bool rv = what?

#endif
                  if(inverted_) rv = !rv;
    
                  return rv;
              }

           case FILEOP_WRITEABLE:
              {
#ifdef _MSC_VER
                  bool rv= !error && (fileStatus.st_mode & S_IWRITE);
#else
                  bool rv= what
#endif
    
                  if(inverted_)
                     rv=!rv;
    
                  return rv;
              }
           case FILEOP_ISDIR:
              {
    
                  bool rv =(!error) && (fileStatus.st_mode & _S_IFDIR);

                  if(inverted_)
                     rv = !rv;

                  return rv;
              }

           case FILEOP_REGULAR:
           {
                   // exists and is not a directory

                   bool rv = (!error) && (0 == (fileStatus.st_mode & _S_IFDIR) );

                   if (inverted_)
                           rv = !rv;

                   return rv;
           }

           case FILEOP_EMPTY:
              {
                  bool rv = error || (fileStatus.st_size == 0);  // no existent files are empty

                                  if (fileStatus.st_mode & _S_IFDIR)
                                          rv = false; // directories don't count

                  if(inverted_)
                     rv = !rv;
                  
                  return rv;
              }

              case FILEOP_NOTEMPTY:
              {
                      // exists, is not a directory, and is not empty

                      bool rv = (!error) || (fileStatus.st_size != 0);

                      if (fileStatus.st_mode & _S_IFDIR)
                              rv = false; // directories don't count


                      if (inverted_)
                              rv = !rv;

                      return rv;
              }

           case FILEOP_EXE:
              {
                  int rv = false;
#ifdef _MSC_VER

                  if(!error && (fileStatus.st_mode & _S_IFDIR))
                  {
                      rv = true; // directories are executable, so true
                  }

                  if(!rv && !error)
                  {
                      size_t lastDot = string_.find_last_of('.');
                      
                      if(lastDot < string_.size())
                      {
                          // there is a filename extension

                          string tmp = cxxtls::StrTool::uppercase(string_.substr(lastDot)); 

                          if(tmp == ".SH" || tmp == ".BRASH")
                          {
                             rv = true;  // yep, executable
                          }

                          char const *pathext = getenv("PATHEXT");
                 
                          if(!rv && pathext)
                          {
                              cxxtls::StrTool::stringlist_t exetypes;
                 
                              cxxtls::StrTool::Is_Delim  functor(';');

                              cxxtls::StrTool::parse_words(pathext,
                                                   &exetypes,
                                                   2000000000U,
                                                   functor
                                                  );
                            
                              cxxtls::StrTool::stringlist_t::iterator a = exetypes.begin(),
                                                                      b = exetypes.end();
                              while(a != b)
                              {
                                  if(*a++ == tmp)
                                  {
                                     rv = true;  // yep, executable
                                     break;
                                  }
                              }
                          }
                      }
                      
                      if (!rv && !error)
                      {


                         std::ifstream x(context->fileHandles_->mapFilename(string_).c_str());
                         
                         std::string line;
                         
                         std::getline(x, line, char(0));
                         
                         if(line.size() > 2 && line[0] == '#' && line[1] == '!')
                         {
                            rv = true;  // yep, executable
                         }

                      }
                  }
#else
                  rv = what?
#endif
                  if(inverted_)
                      rv = !rv;

                  return rv;
              }
              break;

           default:
             throw TestCommandSyntaxInvalidFileOp();  
             break;
       }

       return false;  // fail by default
    }

    virtual string stringValue(ThreadContext* context) const { return string_; }


private:
    string    string_;
    FILEOP    op_;
};

TestExpression *parseTestExpression(std::vector<string>::const_iterator &nextToken, std::vector<string>::const_iterator const &lastToken);

}
