#ifndef CXX_TOOLS_BIscan_header_included_p
#define CXX_TOOLS_BIscan_header_included_p
//
// Copyright 2010, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
#include <string>
#include <string.h> /* for strlen */
#include <stdio.h>  /* for sscanf */

//==========================================================================

///@file BIscan/BIscan.h
/// Defines namespace BIscan which provides routines which perform fast and
/// flexible scanning of character buffers and strings for the builtin types:
///
/// @li integers
/// @li strings
/// @li floats
/// @li character arrays
///

//==========================================================================

/// @page BIscanIntro     A faster sscanf
///
/// namespace BIscan, defined in file @ref BIscan.h, introduces a collection of 
/// text conversion tools that run about 3X the time speed of sscanf when dealing 
/// with integers and which take advantage of C++ template techniques and the 
/// iostream style to make scanning of in-memory buffers easy.
///
/// For example:
///
///@code
///
///  #include <BIscan/BIscan.h>
///
///  ...
///
///  int i1, i2;
///
///  std::string s1;
///
///  char buffer[40];
///
///  if( BIscan::stream("1 2 string moreString") >> i1 >> i2 >> s1 >> buffer )
///  {
///    // scan successful
///  }
///
///@endcode
///
///  Compare the above example with the sscanf variant:
///
///@code
///
///  #include <portable_io.h>
///
///  ...
///
///  int i1, i2;
///
///  std::string s1;  // can't use this!
///
///  char buffer[40];
///
///  char s1_contents[40]; 
///
///  if( 4 == sscanf("1 2 string moreString", &i1, &k2, s1_contents, buffer) )
///  {
///    // scan successful
///
///    s1 = s1_content;
///  }
///
///@endcode
///
///  See @ref BIscan for the list of free functions that perform 
///  various string conversions.
///
///  @note In the above example, BIscan::stream is a function returning 
///        a BIscan::Stream.
///
///  @see BIscan::stream, BIscan::Hex, BIscan::convertInteger

//==========================================================================

namespace BIscan 
    /// Namespace BIscan defines a variety of fast text scanning classes and 
    /// functions which eliminate the overhead found using iostreams 
    /// when scanning simple text buffers for builtin types.  Basically, using 
    /// class @ref BIscan::Stream is like using sscanf -- and is 3X faster for 
    /// integers because of inlining. For example:
    ///
    ///@code
    ///
    ///   int x, y, z;
    ///
    ///   stream("10 11 12") >> x >> y >> z;
    /// 
    ///@endcode
    /// 
    ///
    /// With Stream, there are several operator>>() 
    /// functions that actually trigger the scanning operations, though you 
    /// can call the underlying parsers for even higher speeds if so desired.
    ///
    /// There are many functions in this namespace that do not require the use 
    /// of the Stream class but which let you parse numbers and streams out of 
    /// character buffers directly.  The Stream class is just a thin-wrapper 
    /// around calls to these functions.  Consider:
    ///
    ///   @li convertInteger()
    ///   @li convertFloat()
    ///   @li convertDouble()
    ///   @li parseInteger()
    ///   @li parseUnsignedDecimal()
    ///
    /// Stream doesn't let you parse any old data type of interest -- it is 
    /// focused on integers, floats, char, character arrays, and strings.  
    /// You can if you desire create your own operator>> for your own classes
    /// but the goal of the Stream class is to provide high speed conversion of
    /// builtin types.
    ///
    /// iostreams are very powerful and flexible but also quite slow for this 
    /// particular use.  There is significant overhead found in constructing 
    /// @c stringstream from a character buffer.  For example, creating a 
    /// stringstream and then invoking operator>> to parse an integer is 5 
    /// times slower than using sscanf() and 15 times slower than class Stream's 
    /// operator>>'s.
    ///
    /// BIscan definitions exist to maximize flexibility and minimize overhead 
    /// within the constraints of these limited goals.  Many of the classes are 
    /// designed such that they take iterator references as parameters and can 
    /// thus be used to construct parsers of your own design even if you
    /// choose not to use the higher layers of scanners and converters in this 
    /// namespace. See the functions named parse* and is*.
    ///
    /// @note You can define operator>> for classes of your own design in much 
    ///      the same way that you design them for iostreams.  This can lead to 
    ///      code duplicate though -- one specialization for iostreams and 
    ///      another for Stream -- unless you make your operator>> be templatized 
    ///      on the stream type.
    ///
    /// @warning When you use BIscan::Stream to extract a character, you will 
    ///          get a character not an eight bit signed integer.  An 
    ///          unsigned char, however is treated as a number, not a character.
    ///
{

    // -----------------------------------------------------------------------------------

    inline bool isValidDigit(char c, int base)
        ///
        /// Determine if a character is a valid digit in the specified base.
        ///
        /// @return    true  if c is a valid digit.
        ///
        /// @param[in] c     the character of interest.
        /// @param[in] base  the numeric base of interest.
        ///
        /// @note  Base may only take on values of 2-10 and 16.  Other values
        ///        treated as silent errors and no character will match those bases.
        ///
    {
       if(base == 16)
          return isxdigit(c);
    
       if(base <= 10 && base > 1)
       {
	 return (c >= '0') && ((c - '0') < base);
       }
    
       return false;
    }

    // -----------------------------------------------------------------------------------

    inline int digitValue(char c, int base=10)
        ///
        ///  Determine the numeric value of a character in a given base.  The character
        ///  and the base are assumed to be a valid combination and no error checking
        ///  is done, use isValidDigit() for that.
        ///
        ///  @return      a number in the range 0 - base-1.
        ///
        ///  @param[in]   c     the character to be converted into a number.
        ///  @param[in]   base  the numeric base of the conversion.  (Defaults to 10)
        ///
        ///  @warning  @b base may only take on the following values:  2-10 and 16.
        ///            No error detection is made by this routine.
        ///
    {
       if(base == 16 && isxdigit(c))
       {
          c = toupper(c);
    
          if(c < 'A')
              return c - '0';
    
          return c - 'A' + 10;
    
       }
       else
          return c - '0';
    
       return 0;
    }

    // -----------------------------------------------------------------------------------

    template<class Iterator>
    inline bool parseUnsignedDecimal(Iterator &start, 
                                     Iterator const &end, 
                                     unsigned long long &tmp
                                    )
       /// A hardcoded parser for base 10 numbers which is used to speed it up 
       /// parseInteger() slightly.
       ///
       /// @return     true   if the text found between start and end is a valid unsigned
       ///                    decimal integer.  Note that leading white space is @b NOT 
       ///                    removed prior to the conversion!
       ///
       /// @param[in,out]  start   An interator to the first byte of the digit string.
       /// @param[in]      end     An interator to the first byte not in the string.
       /// @param[out]     tmp     A reference to the output value.
    {
       tmp = 0;
    
       char c;
    
       bool found=false;
    
       while(    start != end 
             &&  ( (c=*start) >= '0')
             &&  ( c <= '9' )
            )
       {
          ++found;
          ++start;
          tmp *= 10;
          tmp += c - '0';
       }
    
       return found;
    
    }
    
    // -----------------------------------------------------------------------------------
    
    
    template<class T, class Iterator>
    inline bool parseInteger(Iterator       &startRef, 
                             Iterator const &end, 
                             T              &integer, 
                             int             base=10, 
                             bool            signable=true
                            )
        ///
        /// Parse a signed or unsigned integer from a character range and store it in a
        /// integer variable.  Remove leading blanks first.
        ///
        /// @return true if the parsing was successfull.
        ///
        /// @param[in,out] startRef     start of character buffer where string resides
        /// @param[in]     end          end of buffer
        /// @param[out]    integer      the output value
        /// @param[in]     base         a number in the range 2-10 or 16.  Defines the base 
        ///                             of the conversion (10 is the default).
        /// @param[in]     signable     a flag indicating that this number can be signed --
        ///                             true by default
        ///
        /// @note:  the input character start value is modified only if the parsing was
        ///         successful.
        ///
    {
       Iterator start = startRef;   // the reference to the start of  buffer does
                                    // not get modified unless parsing is 
                                    // successful.
    
       if(startRef != end)
       {
	   char validator = *startRef; 
                                          // if this line fails to compile, your iterator 
                                          // does not point to a character.
	   validator=validator;
       }
    
       unsigned long long tmp=0;
    
       while(start != end && isspace(*start))
         ++start;
    
       if(start == end)
          return false;
    
       bool negative=false;
    
       if(signable) 
       {
           // signed numbers are allowed
    
           if(*start == '-')
           {
              negative = true;
              ++start;
           }
        
           if(start == end)
              return false;
       }
    
       char c;
    
       if( ! isValidDigit(c = *start, base) )
          return false;
    
       if(base == 10)
       {
           // speed optimization for the most common case
    
           if(! parseUnsignedDecimal(start, end, tmp) )
              return false;
    
       }
       else
       {
           while(    start != end 
                 &&  isValidDigit( c = *start, base )
                )
           {
              ++start;
              tmp *= base;
              tmp += digitValue(c,base);
           }
       }
    
       if(negative)
       {
           long long negValue = tmp;
           negValue = -negValue;
           integer = negValue;
       }
       else
           integer = tmp;
    
       startRef = start; // return the new start of the character buffer.
       return true;
    }

    // -----------------------------------------------------------------------------------

    template<class T>
    bool convertInteger(char const * const &s,  
                        T                  &integer, 
                        int                 base=10, 
                        int                 signable=true
                       )
        ///
        ///  Convert a string, possibly with leading blanks, into an integer. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] integer   a reference to the returned value
        ///  @param[in]  base      the numeric base of the conversion (defaults to 10)
        ///  @param[in]  signable  a flag indicating that the number can be 
        ///                        negative (ie with a leading -) (defaults to true)
        ///
        ///  @note  This function only accepts bases in the range:  2-10 and 16.
        ///
    {
       char const *p = s;  // p is a copy of s, we don't want to pass s to convertInteger
                           // because it modifies its parameter.
    
       return parseInteger(p, p + strlen(s), integer, base, signable);
    }

    // -----------------------------------------------------------------------------------

    template<class T, class StringType>
    bool convertInteger(StringType  const &s,  
                        T                 &integer, 
                        int                base=10, 
                        int                signable=true
                       )
        ///
        ///  Convert a std::string, possibly with leading blanks, into an integer. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] integer   a reference to the returned value
        ///  @param[in]  base      the numeric base of the conversion (defaults to 10)
        ///  @param[in]  signable  a flag indicating that the number can be 
        ///                        negative (ie with a leading -) (defaults to true)
        ///
        ///  @note  This function only accepts bases in the range:  2-10 and 16.
        ///
        ///
    {
       typename StringType::const_iterator tmp(s.begin());  
                                         // convertInteger modifies its first parameter
    
       return parseInteger(tmp, s.end(), integer, base, signable);
    }

    // -----------------------------------------------------------------------------------

    template<class Iterator>
    inline bool parseDouble(Iterator       & firstRef, 
                     Iterator const & last, 
                     double         & d
                    )
        ///
        ///    Parse a double from a string.  Leading blanks are ignored.
        ///
        ///    @return true to indicate success.
        ///
        ///    @param[in,out] firstRef  Defines the start of the string buffer to parse.
        ///    @param[in]     last      Defines the end of the character range.
        ///    @param[out]    d         Specifies the variable to receive the data.
        ///
    {
       Iterator first = firstRef;

       while( first != last && isspace(*first) )
         ++first;

       static std::string floatDigits(".0123456789e+-");

       char buffer[300];

       char *scan = buffer;

       char c;

       while(    first != last 
             &&  floatDigits.find_first_of(c = *first) < floatDigits.size() 
            )
       {
          *scan++ = c;

          ++first;

          if(scan > (buffer+ sizeof(buffer)-1))
             return false;

       }

       if(scan == buffer)
         return false;

       *scan = 0;

       if( 1 != sscanf(buffer, "%lf", &d) )
         return false;

       firstRef = first;
       return true;
    }

    // -----------------------------------------------------------------------------------

    inline bool convertDouble(char const * const &s,  double &d)
        ///
        ///  Convert a string, possibly with leading blanks, into an double. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] d         a reference to the returned value
        ///
    {
       char const *p = s;  // p is a copy of s, we don't want to pass s to convertInteger
                           // because it modifies its parameter.
    
       return parseDouble(p, p + strlen(s), d);
    }

    // -----------------------------------------------------------------------------------

    template<class StringType>
    inline bool convertDouble(StringType const &s,  double &d)
        ///
        ///  Convert a string, possibly with leading blanks, into an double. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] d         a reference to the returned value
        ///
    {
       char const *p = s.data();

       return parseDouble(p, s.data() + s.size(), d);
    }

    // -----------------------------------------------------------------------------------

    inline bool convertFloat(char const * const &s,  float &f)
        ///
        ///  Convert a string, possibly with leading blanks, into an float. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] f         a reference to the returned value
        ///
    {
       char const *p = s;  // p is a copy of s, we don't want to pass s to convertInteger
                           // because it modifies its parameter.  (Only the pointer is copied,
                           // not the data)

       double d;
    
       bool rv = parseDouble(p, p + strlen(s), d);

       f = d;

       return rv; 
    }

    // -----------------------------------------------------------------------------------

    template<class StringType>
    inline bool convertFloat(StringType const &s,  float &f)
        ///
        ///  Convert a string, possibly with leading blanks, into an float. 
        ///
        ///  @return     true      if the conversion is successful
        ///  @param[in]  s         the nul terminated string to convert
        ///  @param[out] f         a reference to the returned value
        ///
    {
       double d;

       char const *p = s.data();

       bool rv = parseDouble(p, s.data() + s.size(), d);

       f = d;

       return rv;
    }

    // -----------------------------------------------------------------------------------



  // forward declare class Stream and various functions that are allowed to
  // construct Stream objects.  Stream objects are meant to be used only as
  // compiler temporaries.  It is not generally allowed to construct named variables
  // of type Stream.  There is little need for it and the Stream::operator!() makes
  // this a bit iffy in practice.  See below.

  class Stream;

  Stream stream(char const * const &p);
  Stream stream(char const * begin, char const *end);

  template<class StringType>
  Stream stream(StringType const &p);

  Stream stream(char const * begin, size_t length);


  class Stream
       /// A Representation of a stream of bytes that can be parsed quickly.  See its
       /// related operator >> functions.  The Stream does not own the bytes, and
       /// so they must not be changed or destroyed while the parsing process is
       /// in effect.
       ///
       /// The purpose of the Stream class is to simulate sscanf using C++ style
       /// syntax and not to create an alternative to the istream class.  Streams are
       /// an extensible form of sscanf in that you can add your own operator>> functions
       /// to do things not originally included in the design -- which is basically 
       /// targetted at numbers and strings.  
       ///
       /// Don't construct objects of this type directly, instead, use the @ref
       /// stream() family of functions to interact with Streams.  Their job is to 
       /// construct streams from parameters that identify the type of the stream.
       ///
       /// At its heart, a stream is a couple of pointers that refer to the as yet
       /// unparsed stream of data.  The operator>> methods do all the work.
       ///
       /// Example use:
       ///
       ///@code
       ///
       ///   std::string somevar, textvar;
       ///   int someInt;
       ///
       ///   if( stream("some 99 text") >> somevar >> someInt >> sometext )
       ///   {
       ///     // Everything is good: somevar contains "some' and textvar contains
       ///     // text.  The variable, someInt contains 99.
       ///     //
       ///   }
       ///   else
       ///   {
       ///     // shouldn't get here because "some 99 text" can be parsed as
       ///     // a string followed by white space followed by an integer followed by a string.
       ///   }
       ///
       ///   if( !stream("glarf") >> someInt )
       ///   {
       ///       // you should get here because "glarf" won't parse to an int.
       ///   }
       ///
       ///@endcode
       ///
       /// @warning When you use BIscan::Stream to extract a character, you will get a character not an
       ///          eight bit signed integer.  An unsigned char, however is treated as a number, not a
       ///          character.
       ///
       /// A Stream is @b NOT an iostream analog.  Rather it is a super simplistic
       /// representation only of a stream of bytes.  It has no internal state --
       /// other than baddness.  Instead of io manipulators, as iostreams use, there are
       /// wrapper classes which allow you to specify the numeric base of the conversion
       /// and or any other filtering like you might want to apply to strings.  Consider
       /// the following wrappers that control the base of conversion:
       /// @li  @ref HexParsed
       /// @li  @ref BinaryParsed
       /// @li  @ref OctalParsed
       ///
       /// These classes are not constructed directly but vi the following corresponding
       /// functions:
       /// @li @ref Hex
       /// @li @ref Binary
       /// @li @ref Octal
       ///
    {

       // Maintenance note:  the design of this class with respect to operator!() and
       // operator void const *(), substituting for operator bool, is very flakey.  
       // Duplicating it would be a bad idea because this class only works because you
       // cannot have variables to class Stream -- you can only have compiler temporaries
       // and that part of the design was necessitated by the goofy way operator! and
       // operator void const *() are required to work.

       char const *start_;
       char const *end_;
       bool        bad_;
       bool        invertedLogic_;  // see operator!() for details.
    
       // -----------------------------------------------------------------------------------
       // the Stream constructor is private and should remain so.
     
       Stream(char const *start, char const *end)
       : start_(start)
       , end_  (end)
       , bad_  (0)
       , invertedLogic_(0)
       {
       }

       // Use a stream() function to construct streams.

       friend Stream stream(char const * const &p);
       friend Stream stream(char const * begin, char const *end);

       template<class StringType>
       friend Stream stream(StringType const &p);


       friend Stream stream(char const * begin, size_t length);


    public:

       // -----------------------------------------------------------------------------------

       operator void const *() const
           /// Return true if no parse errors.  That is, NOT BAD.
       {
           // 
           //  This is a goofy implementation that works only because
           //  class Stream is not allowed to have named variables constructed for it.
           //
           //  See Stream::operator!(), below
           //

           if(invertedLogic_)
              return bad_ ? this : 0;

           return bad_ ? 0 : this; 
       }  

       // -----------------------------------------------------------------------------------

       inline Stream &operator !() 
       { 
           // 
           //  This is a goofy implementation that works only because
           //  class Stream is not allowed to have named variables constructed for it.
           //

           invertedLogic_ ^= true; 
           return *this;  

           // for such a little function, this one needs a lot of comments.
           //
           //  This function would not exist if we only allowed the following
           //  syntaxes for use by Streams:
           //
           //     stream("string") >> v1 >> v2 >> v3;   // uses stream() below to
           //                                           // determine kind of Stream
           //
           //     if( stream("string") >> v4 >> v5 )    // uses stream() below
           //     {
           //        // parsing good!
           //     }
           //
           //
           //  However, if we want the following syntax to work: 
           //
           //    if( !stream("string") >> v1 >> v2 >> v3)
           //    {
           //       // parsing failed!
           //    }
           //
           //  Then we must have this method.
           //
           //  Why?  Good question!  Here's the answer:  Operator ! binds more
           //  strongly than does operator >>.  Thus, '!stream("s")' is evaluated
           //  before the operator>>.  The default operator not (!) returns a bool, and
           //  then operator>> is applied to the bool not to the stream.
           //  Doh!  However, this is the way C++ works.
           //
           //  Why does adding operator! to Stream class solve the problem?  Because
           //  we get to change the return type!  Instead of returning bool, as you
           //  would expect operator! to do, we return the stream itself with a flag set
           //  indicating that its operator! should now work backwards.  Sneaky huh?
           //  
       }
       
       // -----------------------------------------------------------------------------------
    
       bool empty() const { return start_ == end_; } ///< return true if stream empty
    
       char first() const { return *start_; } ///< Returns the first character in
                                              ///  the buffer.  Does not check
                                              ///  for errors.
    
       bool bad() const { return bad_; }      ///< Returns true if a misuse of the
                                              ///  stream has occurred.
    
       void setBad(bool b) { bad_ = b; }      ///< Set or clear the bad flag.
                                              ///  @param b   the new state of the
                                              ///             bad() flag.
    
       void pop_front() { ++start_; }         ///< Discard the first character in
                                              ///  the buffer.
    
       unsigned long size() const { return end_ - start_; }  ///< Return bytes left to be
                                                             ///  parsed.
    
       char const *begin() const { return start_; }   ///< Return iterator to first
                                                      ///  byte to be parsed.
    
       char const *end() const { return end_; }       ///< Return iterator to last
                                                      ///  byte (not part of stream).
    
       // -----------------------------------------------------------------------------------

       template<class IntType>
       Stream &parseInteger(IntType &object, int base=10, bool signable=true)
           // Parse an integer and return it.  Skips leading whitespace.
           // @return      this stream
           // @param[out]  object the data parsed
       {
         if(bad())
           return *this;
    
         if( !BIscan::parseInteger(start_, end_, object, base, signable) )
         {
           setBad(true);
         }
    
         return *this;
       }

       // -----------------------------------------------------------------------------------
    
       Stream &parseDouble(double &d)
           // Parse an double and return it.
           // @return      this stream
           // @param[out]  d the data parsed
       {
         if(bad())
           return *this;
    
         if( !BIscan::parseDouble(start_, end_, d) )
         {
           setBad(true);
         }
    
         return *this;
       }

       // -----------------------------------------------------------------------------------

       void skipWS()
           /// Skip whitespace in the input stream.
       {
          while(start_ != end_ && isspace(*start_))
             ++start_;
       }

       // -----------------------------------------------------------------------------------

       bool skipCharset(char const *first, char const *last)
           ///
           ///  Remove leading characters from the Stream that
           ///  are found in the specified character set.
           ///
           ///  @return    true   to indicate that a character within the
           ///                    character set was actually found.
           ///
           ///  @param[in] first  The first byte of a vector of characters
           ///                    that contains the character set of
           ///                    interest.
           ///  @param[in] last   The last byte of the vector.
           ///
           ///  Leading whitespace is not skipped unless it is contained
           ///  in the character set.
       {
          bool found=false;

          while(start_ != end_)
          {
             char const *scan = first;

             char c = *start_;

             while(scan != last)
             {
                if(c == *scan)
                  break;

		++scan; // didn't match so try the next one
             }

             if(scan == last)
             {
                // current character in the stream is not in the
                // character set
                break;
             }

             found=true;
             ++start_;
          }

          return found;
       }

       // -----------------------------------------------------------------------------------

       bool skipString(char const *s)
          ///
          ///  Remove leading blanks from the input stream, then 
          ///  compare the stream contents with a specified string.
          ///  If there is an exact match, then move the stream
          ///  pointer past the string and return true.
          ///
          ///  Otherwise return false.
          ///
          ///  @return    true   only if the stream contains s.
          ///  @param[in] s      the string to skip over.
          ///
       {
          skipWS();

          char const *start = start_;

          while(start != end_)
          {
             if(*start == *s)
             {
                ++start;
                ++s;
             }
             else
                break;
          }

          if(*s == 0)
          {
            start_ = start;
            return true;
          }

          return false;

       }

       // -----------------------------------------------------------------------------------


       template<size_t N>
       size_t parseString( char (&array)[N] )
           /// Parse a null terminated array of character.
           ///
           /// Skip leading whitespace and parse until the end of 
           /// string or the first whitespace character is found.
           ///
           /// @return the number of bytes read into array (not counting
           ///         the null null byte that will be appended).
           ///
           /// @param[out] array  the array of characters to hold the
           ///                    parsed data.  Note that this must be
           ///                    an array not a pointer!

       {
           if(bad_)
           {
              array[0] = 0;
              return 0;
           }

           skipWS();

           char *output = &array[0];

           size_t avail = N-1;  // leave room for null terminator

           size_t rv=0;

           while(avail-- && start_ != end_ && !isspace(*start_) )
           {
              *output++ = *start_++;
              ++rv;
           }

           *output = 0; // null terminate

            return rv;
       }

       // -----------------------------------------------------------------------------------

       template<class StringType>
       size_t parseString(StringType &s )
           /// Parse text into an stl style string.
           ///
           /// Skip leading whitespace and parse until the end of 
           /// the stream or the first whitespace character is found.
           /// Since the result is a std::string, no nul is appended.
           ///
           /// @return the size of the string parsed.
           ///
           /// @param[out] s  the array of characters to hold the
           ///                parsed data. 

       {
           s.resize(0);

           skipWS();

           char buffer[257];

           for(;;)
           {
               size_t bytesRead = parseString(buffer);

               if(bytesRead == 0)
                  break;

               s.append(buffer, bytesRead);

               if(empty() || isspace(*start_))
                 break;

           }


           return s.size();
       }
    
    }; // class Stream

    // -----------------------------------------------------------------------------------

    template<class T>
    struct IsIntType
        /// Wrapper around the member, value, which will be 1 if T is
        /// an integral type (including unsigned char but not regular
        /// char).  For all other types, value will be 0.
        ///
        /// Yes, this code is in the boost library too -- and probably
        /// works better.  But our needs are low here.
    {
         // none of the following functions and variables is
         // ever defined.  They exist so that sizeof( f(t) )
         // will evaluate to the size of the correct type (either char or int).

         static T* t;  

         static int f( void const * );   // handles all types of data

         static char f(int);
         static char f(short);
         static char f(long);
         static char f(long long);
         static char f(unsigned int);
         static char f(unsigned short);
         static char f(unsigned long);
         static char f(unsigned long long);
         static char f(unsigned char);

         enum { value = (1 == sizeof( f(*t) )) }; 
    };

    // -----------------------------------------------------------------------------------

    template<class T, int TypeCategory=0>
    struct StreamReader
        /// StreamReader is a class used to parse integers out
        /// of stream.  The default implementation will not
        /// allow ANY class to compile.  You must use a specialization
        /// to get the code to compile.   See StreamReader specializations
        /// below.  We made add specializations to StreamReader later to
        /// allow other kinds of classes to be parsable.
    {
    };

    // -----------------------------------------------------------------------------------

    template<class T>
    struct StreamReader<T, 1>
        /// Decimal Integer StreamReader.  
    {
        void operator() (Stream &s, T &outputVariable)
        {
            s.parseInteger(outputVariable);
        }
    };
    
    // -----------------------------------------------------------------------------------

    // specialization 2 MUST NOT EXIST.  See Hex(), below

    // -----------------------------------------------------------------------------------

    template<class T>
    struct StreamReader<T, 3>
        /// Hexadecimal Integer StreamReader.  
    {
        void operator() (Stream &s, T &outputVariable)
        {
            s.parseInteger(outputVariable,16);
        }
    };

    // -----------------------------------------------------------------------------------
    // specialization 4 MUST NOT EXIST.  See Octal(), below

    // -----------------------------------------------------------------------------------

    template<class T>
    struct StreamReader<T, 5>
        /// Octal Integer StreamReader.  
    {
        void operator() (Stream &s, T &outputVariable)
        {
            s.parseInteger(outputVariable,8);
        }
    };

    // -----------------------------------------------------------------------------------

    // specialization 6 MUST NOT EXIST.  See Octal(), below

    // -----------------------------------------------------------------------------------

    template<class T>
    struct StreamReader<T, 7>
        /// Binary Integer StreamReader.  
    {
        void operator() (Stream &s, T &outputVariable)
        {
            s.parseInteger(outputVariable,2);
        }
    };


    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline Stream &operator>> (Stream const & s, IntegerType &variable)
       /// Template operator >> for the Stream class.  It handles
       /// all data types for which specific overloads are not
       /// aleady defined, below. 
       ///
       /// Only integer types are actually supported.
    {
        // s must const & or you can't get his code compile
    
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;
    
      // mystream.parseInteger(variable);

      StreamReader<IntegerType, IsIntType<IntegerType>::value >() (mystream, variable);
         //!  If the above line of code fails to compile, it is because your
         //   variable reference is not an integer.  See IsIntType<> above for a list
         //   of the allowed types.
    
      return mystream;
    }
    
    // -----------------------------------------------------------------------------------

    inline Stream &operator>> (Stream const & s, char &c)
        /// This overload of operator>> allows you to read characters
        /// out of Stream without ignoring whitespace.  Its pretty 
        /// sluggish though, you should probably consider using some
        /// other interface for any kind of bulk processing.  See
        /// @li @ref Stream::first()
        /// @li @ref Stream::empty()
        /// @li @ref Stream::pop_front()
        /// @li @ref Stream::bad()
        /// if you need to process large volumes of data.
    {
        // s must const & or you can't get his code to compile
    
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;
    
      if(mystream.empty())
      {
        mystream.setBad(true);
      }
      else
      {
         c = mystream.first();
    
         mystream.pop_front();
    
      }
    
      return mystream;
    }

#if 0

    J H M C!  The STUPID POS Microsoft compiler does not properly
    generate code for this signature -- sometimes it pretends that it
    did not see it all...  SHEESH

    // -----------------------------------------------------------------------------------

    template<size_t N>
    inline Stream &operator>> (Stream const &s, char (&array)[N] )
        /// Parse a string out of the stream and store it in an
        /// array of characters.  Note that you @b must supply
        /// an array rather than a pointer for this to work.
        ///
        /// Leading blanks are discarded prior to parsing.
        /// 
        /// The output data will be nul terminated so the maximum
        /// string length returned is the size of the array minus
        /// 1.
    {
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;
    
      mystream.parseString(array);
    
      return mystream;
    }
#endif


    // -----------------------------------------------------------------------------------

    inline Stream &operator>> (Stream const &s, char const *skipString)
        ///
        ///  Compare the contents of the stream to a specified string.
        ///  If the stream contents exactly match all the characters in the string,
        ///  in sequence, then the stream is moved passed the match.
        ///
        ///  Matching characters are discarded.
        ///
        ///  @return The stream without the matching string.
        ///
        ///  @param[in,out] s           The stream of interest.
        ///  @param[in]     skipString  The text to match.
        ///
        ///  If the string is not found in the stream, the stream is
        ///  marked as bad.
        ///
        ///  Note:  leading whitespace is skipped in teh stream before
        ///  processing begins.  So a skipString should not contain
        ///  leading whitespace.
        ///
    {
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;
    
      mystream.skipString(skipString);
    
      return mystream;
    }

    // -----------------------------------------------------------------------------------

    inline Stream &operator>> (Stream const &s, std::string &str )
        ///
        /// Parse a std::string out of the Stream.  Skip leading
        /// whitespace.  Does not nul terminate str.
        ///
        /// @return The stream minus the parsed string.
        ///
        /// @param[in]    s    The stream
        /// @param[out]   str  The string to received the parsed data.
        ///
        /// The returned string is set to bad if no string is available.
    {
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;
    
      mystream.parseString(str);
    
      return mystream;
    }

    // -----------------------------------------------------------------------------------

    inline Stream &operator>> (Stream const & s, double &d)
        ///
        /// Skip leading whitespace then parse a double precision floating 
        /// point number from the stream and store it into a double variable.
        ///
        /// @return The stream minus the parsed number.
        ///
        /// @param[in]    s    The stream
        /// @param[out]   d    The double precision floating point value to
        ///                    to received the parsed data.
        ///
        /// The returned string is set to bad if no number is available.
        /// 
    {
    
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;

      mystream.parseDouble(d);

      return mystream;

    }

    // -----------------------------------------------------------------------------------

    inline Stream &operator>> (Stream const & s, float &f)
        /// 
        /// Skip leading blanks then parse a floating point number from 
        /// the stream and store it into a float variable.
        ///
        /// @return The stream minus the parsed number.
        ///
        /// @param[in]    s    The stream
        /// @param[out]   f    The floating point value to
        ///                    to received the parsed data.
        ///
        /// The returned string is set to bad if no number is available.
        /// 
    {
    
      Stream &mystream = const_cast<Stream&>(s);
    
      if(mystream.bad())
         return mystream;

      double d;

      mystream.parseDouble(d);

      if(!mystream.bad())
          f = d;

      return mystream;

    }

    // -----------------------------------------------------------------------------------


    template<class IntegerType>
    struct HexParsed
        ///
        /// This class is a wrapper around a reference to an integer which
        /// exists so that that operator>> can be specialized such that
        /// when a Hex object is serialized, it will parse the stream for a
        /// hex integer instead of a decimal integer.  
        ///
        /// Note:  don't use this class directly -- use the function, @ref Hex(),
        /// below instead.
        ///
        /// Here is an example use:
        ///
        ///@code
        ///
        ///   int value;
        ///
        ///   stream("10") >> BIscan::Hex(value);
        ///
        ///   // value will be 16 (not 10) at this point.
        ///
        ///@endcode
        ///
        /// @note  This works because class there is an overloaded operator >> 
        ///        for class Stream that parses hex integers instead of the 
        ///        default decimal integers.
        ///       
    {
        IntegerType &output_;

        HexParsed(IntegerType &data)
        : output_(data)
        {
        }
    };

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline HexParsed<IntegerType> Hex(IntegerType &t)
        ///
        ///  Construct a wrapper around an integer reference to indicate
        ///  to the Stream parsing logic that the integer should be parsed
        ///  as hexadecimal.
        ///
        ///  @return HexParsed<IntegerType>
        ///
        ///  @param[in,out]  t   The integer to be parsed as hex from the
        ///                      the stream.
        ///
    {
       return HexParsed<IntegerType>(t);
    }

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline Stream &operator>> (Stream const &s,  HexParsed<IntegerType> h )
        ///
        ///  This overload of the Stream output operator is designed
        ///  to trigger hexadecimal rather than decimal parsing of
        ///  of integers from the stream.  The HexParsed<> object
        ///  which is passed in as a parameter is nothing more than
        ///  a wrapper around some integer which should be parsed
        ///  as hex rather than decimal.  Unlike the i/o streams,
        ///  there are no stream states, instead, you accomplish 
        ///  hex parsing like this:
        ///
        ///@code
        ///
        ///   int decimal, int hex;
        ///
        ///   stream("10 10") >> decimal >> Hex(hex);
        ///
        ///   // decimal should be 10 and hex should be 16.
        ///
        ///@endcode
        ///
        ///   @param[in]    s     The stream from which to parse the
        ///                       hex number.
        ///
        ///   @param[out]   h     The HexParsed<> wrapper defining the
        ///                       number to parse as hex.
        ///
        ///   If a hexadecimal number is not found in the stream, the
        ///   returned stream will be marked as bad.
    {
        Stream &myStream = const_cast<Stream&>(s);

        if(myStream.bad())
           return myStream;

        myStream.skipWS();

        if(myStream.empty())
        {
           myStream.setBad(true);
           return myStream;
        }

        StreamReader<IntegerType, (2 + IsIntType<IntegerType>::value) >() (myStream, h.output_);
         //!  If the above line of code fails to compile, it is because your
         //   variable reference is not an integer.  See IsIntType<> above for a list
         //   of the allowed types.

        return myStream;
    }

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    struct OctalParsed
        /// This class is a wrapper around a reference to an integer which
        /// exists so that that operator>> can be specialized such that
        /// when a Octal object is serialized, it will parse the stream for a
        /// octal integer instead of a decimal integer.  
        ///
        /// Note:  don't use this class directly -- use the function, @ref Octal(),
        /// below instead.
        ///
        /// Here is an example use:
        ///
        ///@code
        ///
        ///   int value;
        ///
        ///   stream("10") >> BIscan::Octal(value);
        ///
        ///   // value will be 8 (not 10) at this point.
        ///
        ///@endcode
        ///
        /// @note  This works because class there is an overloaded operator >> 
        ///        for class Stream that parses octal integers instead of the 
        ///        default decimal integers.
        ///       
    {
        IntegerType &output_;

        OctalParsed(IntegerType &data)
        : output_(data)
        {
        }
    };

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline OctalParsed<IntegerType> Octal(IntegerType &t)
        ///
        ///  Construct a wrapper around an integer reference to indicate
        ///  to the Stream parsing logic that the integer should be parsed
        ///  as octal.
        ///
        ///  @return OctalParsed<IntegerType>
        ///
        ///  @param[in,out]  t   The integer to be parsed as octal from the
        ///                      the stream.
        ///
    {
       return OctalParsed<IntegerType>(t);
    }

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline Stream &operator>> (Stream const &s,  OctalParsed<IntegerType> o )
        ///
        ///  This overload of the Stream output operator is designed
        ///  to trigger octal rather than decimal parsing of
        ///  of integers from the stream.  The OctalParsed<> object
        ///  which is passed in as a parameter is nothing more than
        ///  a wrapper around some integer which should be parsed
        ///  as hex rather than decimal.  Unlike the i/o streams,
        ///  there are no stream states, instead, you accomplish 
        ///  hex parsing like this:
        ///
        ///@code
        ///
        ///   int decimal, int octal;
        ///
        ///   stream("10 10") >> decimal >> Octal(octal);
        ///
        ///   // decimal should be 10 and hex should be 8;
        ///
        ///@endcode
        ///
        ///   @param[in]    s     The stream from which to parse the
        ///                       octal number.
        ///
        ///   @param[out]   o     The OctalParsed<> wrapper defining the
        ///                       number to parse as octal.
        ///
        ///   If an octal number is not found in the stream, the
        ///   returned stream will be marked as bad.

    {
        Stream &myStream = const_cast<Stream&>(s);

        if(myStream.bad())
           return myStream;

        myStream.skipWS();

        if(myStream.empty())
        {
           myStream.setBad(true);
           return myStream;
        }

        StreamReader<IntegerType, (4 + IsIntType<IntegerType>::value) >() (myStream, o.output_);
         //!  If the above line of code fails to compile, it is because your
         //   variable reference is not an integer.  See IsIntType<> above for a list
         //   of the allowed types.

        return myStream;
    }

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    struct BinaryParsed
        /// This class is a wrapper around a reference to an integer, which
        /// exists so that that operator>> can be specialized such that
        /// when a binary object is serialized, it will parse the stream for a
        /// binary integer instead of a decimal integer.  
        ///
        /// Note:  don't use this class directly -- use the function, @ref Binary(),
        /// below instead.
        ///
        /// Here is an example use:
        ///
        ///@code
        ///
        ///   int value;
        ///
        ///   stream("10") >> BIscan::Binary(value);
        ///
        ///   // value will be 2 (not 10) at this point.
        ///
        ///@endcode
        ///
        /// @note  This works because class there is an overloaded operator >> 
        ///        for class Stream that parses binary integers instead of the 
        ///        default decimal integers.
        ///       

    {
        IntegerType &output_;

        BinaryParsed(IntegerType &data)
        : output_(data)
        {
        }
    };

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline BinaryParsed<IntegerType> Binary(IntegerType &t)
        ///
        ///  Construct a wrapper around an integer reference to indicate
        ///  to the Stream parsing logic that the integer should be parsed
        ///  as binary.
        ///
        ///  @return BinaryParsed<IntegerType>
        ///
        ///  @param[in,out]  t   The integer to be parsed as binary from the
        ///                      the stream.
        ///
    {
       return BinaryParsed<IntegerType>(t);
    }

    // -----------------------------------------------------------------------------------

    template<class IntegerType>
    inline Stream &operator>> (Stream const &s,  BinaryParsed<IntegerType> o )
        ///
        ///  This overload of the Stream output operator is designed
        ///  to trigger binary rather than decimal parsing of
        ///  of integers from the stream.  The BinaryParsed<> object
        ///  which is passed in as a parameter is nothing more than
        ///  a wrapper around some integer which should be parsed
        ///  as hex rather than decimal.  Unlike the i/o streams,
        ///  there are no stream states, instead, you accomplish 
        ///  hex parsing like this:
        ///
        ///@code
        ///
        ///   int decimal, int binary;
        ///
        ///   stream("10 10") >> decimal >> BIscan::Binary(binary);
        ///
        ///   // decimal should be 10 and hex should be 2;
        ///
        ///@endcode
    {
        Stream &myStream = const_cast<Stream&>(s);

        if(myStream.bad())
           return myStream;

        myStream.skipWS();

        if(myStream.empty())
        {
           myStream.setBad(true);
           return myStream;
        }

        StreamReader<IntegerType, (6 + IsIntType<IntegerType>::value) >() (myStream, o.output_);

         //  If the above line of code fails to compile, it is because your
         //   variable reference is not an integer.  See IsIntType<> above for a list
         //   of the allowed types.

        return myStream;
    }

    // -----------------------------------------------------------------------------------

    struct SKIPSET
        /// An object which represents a set of characters which a Stream should ignore when
        /// parsing with operator>>.  The parsed characters are discarded from the stream.
        /// At least one member of the set must appear or an error results.
        ///
        ///@code
        ///
        ///   int x, y;
        ///
        ///   stream("19 , 20") >> x >> SKIPSET(", ") >> y;
        ///
        ///   // here, x will be 19 and and y will be 20 because the SKIPSET specifies that
        ///   // space and comma re ignorable characters.  However, there must be a match
        ///   // or the parsing will stop.
        ///@endcode
        ///
    {
        char const *first_;
        char const *last_;

        SKIPSET(char const *f, char const *l)
        : first_(f)
        , last_(l)
        {
        }

        explicit SKIPSET(char const *s)
        : first_(s)
        , last_( s + strlen(s))
        {
        }

    };

    // -----------------------------------------------------------------------------------

    inline Stream &operator>>( Stream &s, SKIPSET const &cs)
        ///
        ///  Verify that the specified stream contains one or more characters which
        ///  is found in the specified character set.
        ///
        ///  Consume all matching characters from the stream.  The stream must contain
        ///  at least one matching character or the conversion is determined to have
        ///  failed.
        ///
        ///  Leading whitespace is not skipped unless defined in the character set.
        ///
        ///  @return              A stream which no longer contains matching characters.
        ///  
        ///  @param[in,out]   s   The stream of interest.
        ///  @param[in]       cs  The characterset to match against.
        ///
        ///  Note to skip whitespace, declare a character set containing " \r\n\t\b\a"
    {
       if(! s.skipCharset(cs.first_, cs.last_) )
       {
         s.setBad(true);
       }

       return s;
    }

    // -----------------------------------------------------------------------------------

    inline Stream stream(char const * const &p)
        /// One of a family of overloaded functions that returns a 
        /// stream whose type is based on its parameters.
        ///
        /// The Stream returned does not have a copy of the stream
        /// data, only a pointer to it.  You must not destruct or change
        /// the input pointer's data until parsing is complete.
        ///
        /// This function converts a C style string into a Stream.
    {
       return Stream(p,p + strlen(p));
    }

    // -----------------------------------------------------------------------------------

    inline Stream stream(char const * begin, char const *end)
        /// One of a family of overloaded functions that returns a 
        /// stream whose type is based on its parameters.
        ///
        /// The Stream returned does not have a copy of the stream
        /// data, only a pointer to it.  You must not destruct or change
        /// the input pointer's data until parsing is complete.
        ///
        /// This function converts a character byte range into a Stream

    {
       return Stream(begin,end);
    }

    // -----------------------------------------------------------------------------------

     inline Stream stream(char const * begin, size_t length)
        /// One of a family of overloaded functions that returns a 
        /// stream whose type is based on its parameters.
        ///
        /// The Stream returned does not have a copy of the stream
        /// data, only a pointer to it.  You must not destruct or change
        /// the input pointer's data until parsing is complete.
        ///
        /// This function converts a character byte range into a Stream

    {
       return Stream(begin,begin + length);
    }

    // -----------------------------------------------------------------------------------

    template<class StringType>
    inline Stream stream(StringType const &s)
        /// One of a family of overloaded functions that returns a 
        /// stream whose type is based on its parameters.
        ///
        /// The Stream returned does not have a copy of the stream
        /// data, only a pointer to it.  You must not destruct or change
        /// the input pointer's data until parsing is complete.
        ///
        /// This function converts an stl style string into a Stream.
    {
       return Stream( s.data(), s.data() + s.size() );
    }


} // namespace BIscan

#endif
