#ifndef ArithmeticExpressionParser_h_included_p
#define ArithmeticExpressionParser_h_included_p
//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#include <cxxtls/stringexpression.h>
#include <cxxtls/lexical_cast.h>


///@file
///  This file defines the parser for the BASH arithmetic expression
///  syntax as used by the let command and the (( expr )) syntax.
///


namespace cxxtls
{
namespace StringExpr
{
namespace Arithmetic
{


struct Node
    /// a node in an arithmetic expression tree
{
    virtual long execute(ExecutionContext *context) const=0;
    virtual ~Node() {}
};

struct    IntegerLiteral
: public  Node
    ///   a node in an arithmetic expression tree that evaluates to a constant value.
{
    long value_;

    IntegerLiteral(long value=0)
    :  value_(value)
    {
    }

    virtual long execute(ExecutionContext*) const { return value_; }
};

struct    Variable
: public  Node
    ///   a node in an arithmetic expression tree that evaluates to a variable's
    ///   integer value.
{
    std::string name_;

    Variable(std::string const &name)
    :  name_(name)
    {
    }

    virtual long execute(ExecutionContext *context) const 
    {
        long rv;

        try { rv = lexical_cast<long>(context->find(name_)); }
        catch (...)
        {
            rv=0;
        }
        return rv;
    }
};

struct Not
: public  Node
    ///   a node in an arithmetic expression tree that evaluates to the logical inversion
    ///   of the value on the right hand side.
{
    std::auto_ptr<Node> rhs_;

    Not(Node *rhs)
    :  rhs_(rhs)
    {
    }

    virtual long execute(ExecutionContext *context) const 
    {
        return ! (rhs_->execute(context) );
    }

};

struct Negate
: public  Node
    ///   a node in an arithmetic expression tree that evaluates to "0 - rightHandSideValue"
{
    std::auto_ptr<Node> rhs_;

    Negate(Node *rhs)
    :  rhs_(rhs)
    {
    }

    virtual long execute(ExecutionContext *context) const 
    {
        return (0 - rhs_->execute(context) );
    }
};


struct BitwiseInvert
: public  Node
    ///   a node in an arithmetic expression tree that evaluates to a bitwise inversion
    ///   of the right hand side value.
{
    std::auto_ptr<Node> rhs_;

    BitwiseInvert(Node *rhs)
    :  rhs_(rhs)
    {
    }

    virtual long execute(ExecutionContext *context) const 
    {
        return ~ rhs_->execute(context);
    }
};


#define DECLARE_ARITH_BINOP(name, operation) \
struct name : public Node \
{ \
     std::auto_ptr<Node> lhs_; \
     std::auto_ptr<Node> rhs_; \
     \
     name(Node *lhs, Node* rhs) \
     : lhs_(lhs) \
     , rhs_(rhs) \
     {\
     }\
     \
    virtual long execute(ExecutionContext *context) const \
    { \
        return (lhs_->execute(context)) operation (rhs_->execute(context)); \
    }\
}


DECLARE_ARITH_BINOP(Add,       +);
DECLARE_ARITH_BINOP(Subtract,  -);

DECLARE_ARITH_BINOP(Multiply,  *);
DECLARE_ARITH_BINOP(Divide,    /);
DECLARE_ARITH_BINOP(Modulo,    %);

DECLARE_ARITH_BINOP(Greater,   >);
DECLARE_ARITH_BINOP(Less,      <);
DECLARE_ARITH_BINOP(Equal,    ==);
DECLARE_ARITH_BINOP(NEqual,   !=);
DECLARE_ARITH_BINOP(GEqual,   >=);
DECLARE_ARITH_BINOP(LEqual,   <=);

DECLARE_ARITH_BINOP(Xor,      ^);
DECLARE_ARITH_BINOP(And,      &);
DECLARE_ARITH_BINOP(Or,       |);


DECLARE_ARITH_BINOP(AndAnd,   &&);
DECLARE_ARITH_BINOP(OrOr,     ||);


struct Assignment
    ///   a node structure representing an arithmetic assignment.  It has
    ///   a variable name and an Arithmetic Expression.
{
    std::string           name_;
    std::auto_ptr<Node>   value_;

    Assignment(std::string const &varname, Node *value)
    :  name_(varname)
    ,  value_(value)
    {
    }

    long execute(StringExpr::ExecutionContext *context) const
    {
        long value = value_->execute(context);


        if(!name_.empty())
        {
            std::string value_rep = lexical_cast<std::string>(value);

            if(context->logger_ && context->logging())
            {
               std::string message=name_ + "=" + value_rep;

               context->logger_->log(message);
            }

            context->assign(name_, value_rep);
        }

        return value;
    }

};


struct PreIncrement
    /// increment a variable and return the new value
: public Node
{
    std::string name_;

    PreIncrement(std::string const &varname)
    :   name_(varname)
    {
    }

    long execute(StringExpr::ExecutionContext *context) const
    {

        if(name_.empty())
           return 1;


        long value;
        
        try
        {
            value = lexical_cast<long>(context->find(name_));
        }
        catch(...)
        {
            value = 0;
        }

        ++value;


        context->assign(name_, lexical_cast<std::string>(value));

        return value;
    }

};

struct PreDecrement
    /// decrement a variable and return the new value
: public Node
{
    std::string name_;

    PreDecrement(std::string const &varname)
    :   name_(varname)
    {
    }

    long execute(StringExpr::ExecutionContext *context) const
    {

        if(name_.empty())
           return -1;

        long value;
        
        try
        {
            value = lexical_cast<long>(context->find(name_));
        }
        catch(...)
        {
            value=0;
        }

        --value;


        context->assign(name_, lexical_cast<std::string>(value));

        return value;
    }

};

struct PostIncrement
    /// increment a variable and return the new value
: public Node
{
    std::string name_;

    PostIncrement(std::string const &varname)
    :    name_(varname)
    {
    }

    long execute(StringExpr::ExecutionContext *context) const
    {

        if(name_.empty())
           return 1;

        long value;
        
        try
        {
            value = lexical_cast<long>(context->find(name_));
        }
        catch(...)
        {
            value = 0;
        }

        context->assign(name_, lexical_cast<std::string>(value+1)); // increment occurs here

        return value;
    }

};

struct PostDecrement
    /// decrement a variable and return the old value
: public Node
{
    std::string name_;

    PostDecrement(std::string const &varname)
    :    name_(varname)
    {
    }

    long execute(StringExpr::ExecutionContext *context) const
    {

        if(name_.empty())
           return -1;

        long value;
        
        try
        {
            value = lexical_cast<long>(context->find(name_));
        }
        catch(...)
        {
            value = 0;
        }

        context->assign(name_, lexical_cast<std::string>(value-1));   // decrement occurs here

        return value;
    }

};





template <class Iterator1, class Iterator2>
inline void skipSeparators(Iterator1 &f, Iterator2 const &l)
    /// move iterator f past any leading spaces
{

   while(f != l && isspace(*f)) ++f;
    
}

template <class Iterator1, class Iterator2>
bool parseVariableName(Iterator1 &f, Iterator2 const &l, std::string &name)
    /// parse a name from a substring.  Move the first iterator
    /// past the name.
    ///
    /// @return true if found
    /// @return name will be populated with the discovered name
{

    name.resize(0);

    skipSeparators(f,l);

    bool firstCharState=true;

    while(f != l)
    {
       char c = *f;

       if(firstCharState)
       {
          firstCharState=false;

          if(!(c == '_' || isalpha(c)))
          {
             return false;
          }
       }

       if(!(c == '_' || isalnum(c)))
       {
          break;
       }

       name += c;
       ++f;

    }

    return  !name.empty();
}

template <class Iterator1, class Iterator2>
Node* parseExpression(Iterator1 &f, Iterator2 const &l);


template <class Iterator1, class Iterator2>
Node *parsePrimary(Iterator1 &f, Iterator2 const &l)
   /// parse expression factors and construct nodes that
   /// will multiply, divide, or modulo them as needed
   ///
   /// The caller owns the return node pointer and must delete it.
   /// 
   /// A primary looks like one of the following syntaxes
   ///
   ///begincode
   ///
   ///    1234
   ///    var
   ///    !expression
   ///    -expression
   ///    (expression)
   ///    ++var
   ///    --var
   ///    var++
   ///    var--
   ///
   ///endcode
{
    std::auto_ptr<Node> expression;

    skipSeparators(f,l);

    if(f == l)
       return 0;

    switch(*f)
    {
         case '!':  
         {
             ++f;

             expression.reset(parseExpression(f,l));

             if(!expression.get())
                return 0;

             std::auto_ptr<Node> subexp;

             {
                 // keep together for exception saftey
                 subexp.reset( new Not(expression.get()) );
                 expression.release();
             }

             {
                // keep together for exception saftey
                expression.reset(subexp.get());
                subexp.release();
             }


         }
         break;

         case '+':
         {
            ++f;

            if(f != l && *f == '+')
            {
                 std::string name;

                 ++f;

                 if(!parseVariableName(f, l, name))
                    return 0;  // error message will name the current line

                 expression.reset( new PreIncrement(name) );
            }
            else
               return 0; // + used out of place

         }
         break;

         case '-':  
         {
             ++f;

             if(f != l && *f == '-')
             {
                 // parse a predecrement

                 ++f;

                 std::string name;

                 if(!parseVariableName(f, l, name))
                    return 0;  // error message will name the current line

                 expression.reset( new PreDecrement(name) );

             }
             else
             {
                 // parse a negation
    
                 expression.reset(parseExpression(f,l));
    
                 if(!expression.get())
                    return 0;
    
                 std::auto_ptr<Node> subexp;
    
                 {
                     // keep together for exception saftey
                     subexp.reset( new Negate(expression.get()) );
                     expression.release();
                 }
    
                 {
                    // keep together for exception saftey
                    expression.reset(subexp.get());
                    subexp.release();
                 }
    
             }


         }
         break;

         case '~':  
         {
             ++f;

             expression.reset(parseExpression(f,l));

             if(!expression.get())
                return 0;

             std::auto_ptr<Node> subexp;

             {
                 // keep together for exception saftey
                 subexp.reset( new BitwiseInvert(expression.get()) );
                 expression.release();
             }

             {
                // keep together for exception saftey
                expression.reset(subexp.get());
                subexp.release();
             }


         }
         break;

         case '(':  
         {
             ++f;

             expression.reset(parseExpression(f,l));

             if(!expression.get())
                return 0;

             skipSeparators(f,l);

             if(f == l)
                return 0;

             if(*f != ')')
                return 0;

             ++f;

         }
         break;

         default:
             if(isdigit(*f))
             {
                 long value=0;

                 while(f != l && isdigit(*f))
                 {
                    value *= 10;
                    
                    value += (*f++ - '0');
                 }

                 expression.reset( new IntegerLiteral(value) );

             }
             else
             {
                 std::string name;

                 if(!parseVariableName(f, l, name))
                    return 0;

                 skipSeparators(f,l);

                 if(f !=l && *f == '+')
                 {
                    // check for var++

                    Iterator1 next = f; ++next;

                    if(next != l && *next == '+')
                    {
                        ++f;
                        ++f;

                        expression.reset( new PostIncrement(name) );
                    }
                    else
                     expression.reset( new Variable(name) );


                 }
                 else
                 if(f !=l && *f == '-')
                 {
                    // check for var--

                    Iterator1 next = f; ++next;

                    if(next != l && *next == '-')
                    {
                        ++f;
                        ++f;

                        expression.reset( new PostDecrement(name) );
                    }
                    else
                     expression.reset( new Variable(name) );

                 }
                 else
                     expression.reset( new Variable(name) );
             }
         break;
    }

    return expression.release();
}


template <class Iterator1, class Iterator2>
Node *parseFactor(Iterator1 &f, Iterator2 const &l)
   /// parse expression factors and construct nodes that
   /// will multiply, divide, or modulo them as needed
   ///
   /// The caller owns the return node pointer and must delete it.
   ///
   ///
   /// 
{
    std::auto_ptr<Node>  expression( parsePrimary(f,l) );

    if(expression.get())
    {
       for(;;)
       {
           skipSeparators(f,l);
           
           if(f == l)
              break;

           char op = *f;

           if(op == '*' || op == '/' || op == '%' )
           {
              ++f;

              std::auto_ptr<Node> rhs( parsePrimary(f,l) );

              if(rhs.get() == 0)
              {
                 // failure to parse the right hand side

                 return 0;  
              }

              if(op == '*')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Multiply(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '/')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Divide(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '%')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Modulo(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
                 return 0; // huh?

           }
           else
              break;
       }
    }
    else
        return 0;

    return expression.release();
}



template <class Iterator1, class Iterator2>
Node * parseTerm(Iterator1 &f, Iterator2 const &l)
   /// parse expression terms and construct nodes that
   /// will add or subtract them as needed.
   ///
   /// The caller owns the return node pointer and must delete it.
   /// 
{
    std::auto_ptr<Node>  expression( parseFactor(f,l) );

    if(expression.get())
    {
       for(;;)
       {
           skipSeparators(f,l);
           
           if(f == l)
              break;

           char op = *f;

           if(op == '+' || op == '-')
           {
              ++f;

              std::auto_ptr<Node> rhs( parseFactor(f,l) );

              if(rhs.get() == 0)
              {
                 // failure to parse the right hand side

                 return 0;  
              }

              if(op == '+')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Add(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '-')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Subtract(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
                 return 0; // huh?

           }
           else
              break;
       }
    }
    else
        return 0;

    return expression.release();

}


template <class Iterator1, class Iterator2>
Node * parseComparison(Iterator1 &f, Iterator2 const &l)
   /// parse expression terms and construct nodes that
   /// will add or subtract them as needed.
   ///
   /// The caller owns the return node pointer and must delete it.
   /// 
{
    std::auto_ptr<Node>  expression( parseTerm(f,l) );

    if(expression.get())
    {
       for(;;)
       {
           skipSeparators(f,l);
           
           if(f == l)
              break;

           char op = *f;

           bool addEqual = false;

           if(op == '>' || op == '<' || op == '=' || op == '!')
           {
              ++f;

              if(f != l)
              {
                if(*f == '=')
                {
                   addEqual = true;
                   ++f;
                }
              }

              if(op == '=' && !addEqual)
                 return 0;   // invalid syntax, you can't use assignment operator here


              std::auto_ptr<Node> rhs( parseTerm(f,l) );

              if(rhs.get() == 0)
              {
                 // failure to parse the right hand side

                 return 0;  
              }

              if(op == '<')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     if(addEqual)
                         subexp.reset(new LEqual(expression.get(), rhs.get()));
                     else
                         subexp.reset(new Less(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '>')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     if(addEqual)
                         subexp.reset(new GEqual(expression.get(), rhs.get()));
                     else
                         subexp.reset(new Greater(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '!')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     if(addEqual)
                         subexp.reset(new NEqual(expression.get(), rhs.get()));
                     else
                         return 0;

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '=')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Equal(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
                 return 0; // huh?

           }
           else
              break;
       }
    }
    else
        return 0;

    return expression.release();

}

template <class Iterator1, class Iterator2>
Node * parseBitwise(Iterator1 &f, Iterator2 const &l)
   /// parse expression terms and construct nodes that
   /// will add or subtract them as needed.
   ///
   /// The caller owns the return node pointer and must delete it.
   /// 
{
    std::auto_ptr<Node>  expression( parseComparison(f,l) );

    if(expression.get())
    {
       for(;;)
       {
           skipSeparators(f,l);
           
           if(f == l)
              break;

           char op = *f;

           bool addEqual = false;   // not equal in this case, but rather duplicate:  && or ||

           if(op == '^' || op == '&' || op == '|')
           {
              ++f;

              if(f != l)
              {
                if(*f == op)
                {
                   addEqual = true;
                   ++f;
                }
              }

              if(op == '^' && !addEqual)
                 return 0;   // invalid syntax, you can't use ^^


              std::auto_ptr<Node> rhs( parseComparison(f,l) );

              if(rhs.get() == 0)
              {
                 // failure to parse the right hand side

                 return 0;  
              }

              if(op == '^')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     subexp.reset(new Xor(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '|')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     if(addEqual)
                         subexp.reset(new OrOr(expression.get(), rhs.get()));
                     else
                         subexp.reset(new Or(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
              if(op == '&')
              {
                  std::auto_ptr<Node> subexp;

                  {
                     // keep together for exception saftey

                     if(addEqual)
                         subexp.reset(new AndAnd(expression.get(), rhs.get()));
                     else
                        subexp.reset(new And(expression.get(), rhs.get()));

                     expression.release();
                     rhs.release();
                  }

                  {
                     // keep together for exception saftey

                     expression.reset(subexp.get());
                     subexp.release();
                  }
              }
              else
                 return 0; // huh?

           }
           else
              break;
       }
    }
    else
        return 0;

    return expression.release();

}

template <class Iterator1, class Iterator2>
Node * parseExpression(Iterator1 &f, Iterator2 const &l)
{
   return parseBitwise(f,l);
}


template <class Iterator1, class Iterator2>
Assignment* parseAssignment(Iterator1 &f, Iterator2 const &l)
   /// parse a variable assignment out of a string.
   ///
   /// The assignments look like  "name=valueExpression" where the name
   /// is defined rules that variables must obey in stringexpression.h.
   ///
   /// The values can be numeric literals, the names of other variables, and operations on
   /// the literals and values as defined by classes above.
   ///
   /// Note that the parser does not evaluate $values.  This must be done before invoking
   /// the parser.
   ///
   /// @return null if there is a syntax error.
   ///
   /// @note:  the caller must delete the return pointer to prevent memory leaks.
{
    std::auto_ptr<Assignment> rv;

    std::string resultVarname;

    std::string::const_iterator startF = f;

    if(parseVariableName(f,l, resultVarname))
    {
      skipSeparators(f,l);
      
      if(f != l && *f == '=')
      {
         ++f;

         if(f != l && *f == '=')
         {
            // we do not have an assignment statement, we have a comparison

            f = startF;

            resultVarname.resize(0);

         }
         
         
         std::auto_ptr<Node> expression(parseExpression(f,l));
         
         if(expression.get())
         {
            {
                // keep thse lines together for exception saftey.
                rv.reset( new Assignment(resultVarname, expression.get()) );
                expression.release();
            }
         }
         else
            return 0;
         
      }
      else
      {
         // we started with a variable name but found no equal sign after it
         // treat this is as just an expression to be evaluated

         resultVarname.resize(0);
       
         f = startF;
       
          std::auto_ptr<Node> expression(parseExpression(f,l));
       
          if(expression.get())
          {
             {
                 // keep thse lines together for exception saftey.
                 rv.reset( new Assignment(resultVarname, expression.get()) );
                 expression.release();
             }
          }
          else
             return 0;

      }
    }
    else
    {
        // we did not begin with variable name, so lets call this a simple expression, not an assignment

        resultVarname.resize(0);

        f = startF;

         std::auto_ptr<Node> expression(parseExpression(f,l));

         if(expression.get())
         {
            {
                // keep thse lines together for exception saftey.
                rv.reset( new Assignment(resultVarname, expression.get()) );
                expression.release();
            }
         }
         else
            return 0;

    }

    skipSeparators(f,l);

    if(f!=l)
    {
       return 0;  // some for of trailing syntax error.
    }

    return rv.release();
}





} // Arithmetic
} // StringExpr
} // cxxtls

#endif
