//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.

#ifndef cxxtls_scriptfunctions_header_included_p
#define cxxtls_scriptfunctions_header_included_p

///@file
///   This file defines script language functions and statements and the mechanism
///   for storage thereof.  Some basic types of statements are included:
///
///   @li variable assignments
///   @li block statemtents (i.e. statement lists)
///   @li function definition statements
///

#include <cxxtls/scriptthread.h>
#include <cxxtls/foreach.h>
#include <cxxtls/threadSafeHandle.h>

namespace cxxtls
{
namespace ScriptFunctions
{
    enum FunctionReturnCode
    {
       FRC_Success,  // function returned with error code of 0.
       FRC_Fail,     // function returned with a non-zero error code, see
                     // executionContext for details
       FRC_Exception // function through an exception, see executionContext for details
    };

    enum StatementReturnCode
    {
       SRC_Success,     // normal return.  (does not mean "exit code == 0"!  just means normal statement execution)
       SRC_Return,      // statement was a return statement, see executionContext for rv
       SRC_Throw,       // statement was a throw statement, see executionContext for type
       SRC_Break,       // statement was "break"
       SRC_Continue,    // statement was "continue"
    };

    struct FunctionScope;
    struct Statement;

    struct JobController
        /// A base class for all kinds of various execution systems.  
        /// The idea is that you derive your execution system from JobController
        /// so that during execute() of statements, they can cast the JobController pointer
        /// to be the actual execution system type, then implement execution system
        /// specific behavior.
    {
        virtual ~JobController()
        {
        }

        virtual StatementReturnCode executeDetached(ScriptThread::ThreadContext *executionContext,
                                                    FunctionScope                *functionScope,
                                                    JobController                *jobController,
                                                    ThreadSafeHandle<Statement>   &statement
                                                   ) 
        {
           return SRC_Throw;  // you must superseed this if you want background jobs to work.
        }
    };


    struct Statement
        /// A representation of a compiled statement in a script language
    {
        virtual ~Statement();

        bool detached_;

        Statement()
        :  detached_(false)
        {
        }

        void setDetached(bool v = true)
        {
          detached_  = v;
        }

        bool detached() const
        {
           return detached_;
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                            FunctionScope                *functionScope,
                                            JobController                *jobController,
                                            bool                          skipArg0 = false    // for use by the brash command "builtin"
                                           ) const = 0;
            ///<  Executing this function in the context of executionContext
            ///   using the specified parameters.
            ///
            ///   @return A command code that indicates what happened in the function.
            ///


        virtual std::string getDebugDescription() const = 0;  // get a string roughly describing the statement
    };



    struct VariableAssignmentStatement
    : public Statement
    {
        virtual ~VariableAssignmentStatement() {}

        std::string varname_;
        std::auto_ptr<cxxtls::StringExpression> value_;
        std::auto_ptr<cxxtls::StringExpression> indexExpression_; // assuming value_ is an array or map


        VariableAssignmentStatement(std::string varname, cxxtls::StringExpression *e)
        :  varname_(varname)
        ,  value_(e)
        {
        }

        virtual StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                            FunctionScope                *functionScope,
                                            JobController                *jobController,
                                            bool                          skipArg0            // for use by the command "builtin"
                                           ) const;
           /// Evaluates value_ in the context of executionContext and sets the variable whose
           /// name is in varname_ in executionContext with that value.
           ///
           /// Note that a search is done within the function scope for the specified varname_.
           /// If any parent scope has a variable of the specified name, then that is the variable
           /// that is modified by this statement.  If no parent has the variable, then a new one
           /// is added to the current scope.
           ///
           /// @return SRC_Success

        virtual std::string getDebugDescription() const 
        {
            /// get a string roughly describing the statement

            return varname_ + "?";
        }

    };


    struct VariableDefinitionStatement
    : public Statement
    {
        virtual ~VariableDefinitionStatement();
        std::string varname_;
        StringExpression value_;

        virtual StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                            FunctionScope                *functionScope,
                                            JobController                *jobController,
                                            bool                          skipArg0            // for use by the command "builtin"
                                           ) const;
           /// Evaluates value_ in the context of executionContext and sets the variable whose
           /// name is in varname_ in executionContext with that value.
           ///
           /// This statement only checks for the current function scope to see if a a variable
           /// has already been defined or not.  It does not search the parent.  Thus it has
           /// the effect of guaranteeing the creation of the variable name in the current
           /// scope.
           ///
           /// @return SRC_Success

        virtual std::string getDebugDescription() const 
        {
            /// get a string roughly describing the statement

            return varname_ + "?";
        }

    };


    class StatementList
        /// A representation of a compiled list of Statements which will be executed in sequence
        /// using the specified executionContext, functionScope.  The command line arguments are
        /// assumed to already be populated in the executionContext
    {
        std::list< ThreadSafeHandle<Statement> > code_;
        
        bool subShell_;   ///<  A flag that indicates that this statement list is meant to be run in a sub-shell.
                          ///   The execution system may or may not wait on the child shell to finish but the
                          ///   environment variables modified in the subshell will not affect the parent
                          ///   thread's environment.

        bool alreadyDetached_;   ///< A flag that means this statement list was created for the purpose of
                                 ///  running a detached job.  If true, then "detached" jobs in this list
                                 ///  will run normally -- this is needed to prevent infinite recursion.

    public:

        StatementList()
        :  subShell_(false)
        ,  alreadyDetached_(false)
        {
        }

        void setAlreadyDetached(bool v)
        {
           alreadyDetached_ = v; // only used by JobController::executeDetached()
        }

        bool subShellFlag() const 
            ///  get the subshell flag.  @see setSubShellFlag
        { 
            return subShell_; 
        }

        void setSubShellFlag(bool flag=true) 
           ///   Set the subshell flag for this statement list.  When true, the environment variables
           ///   are copied for this statement list rather than being shared with parent thread's 
           ///   variables, as is the default.
           ///
           ///   @see subShell_ above.
        {
           subShell_ = flag;
        }

        void push_back(ThreadSafeHandle<Statement> &s)
            /// take ownership of a new statement and put it at the end of the list
        {
           ThreadSafeHandle<Statement> tmp;

           code_.push_back(tmp);

           code_.back() = s;
        }

        void pop_back()
        {
           code_.pop_back();
        }

        ThreadSafeHandle<Statement> back()
        {
           return code_.back();
        }

        size_t size() const
            /// Get t
        {
           return code_.size();
        }

        bool empty() const
        {
           return code_.empty();
        }

        std::list< ThreadSafeHandle<Statement> > const &statements()  const
        {


           return code_;
        }


        StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                    FunctionScope *scope,
                                    JobController                *jobController,
                                    bool                          skipArg0            // for use by the command "builtin"
                                   ) const;
            /// execute the statements in the specified code_ list and handle the return codes
            /// of each.  If nothing wierd happens, return SRC_Success.

        virtual std::string getDebugDescription() const
        {
           std::string rv("{ ");

           int count=0;

           CXXTLS_FOREACH( ThreadSafeHandle<Statement> const &cur, code_ )
           {
              ++count;

              if(count > 1)
                  rv += " ; ";

              rv +=  cur->getDebugDescription();


              if(count > 5)
              {
                 rv += " ... ";
                 break;
              }
           }

           rv += " }";

           return rv;
        }

    private:
        StatementList(StatementList const &);  // not implemented
        StatementList &operator=(StatementList const &); // not implemented
    };

    struct StatementListStatement
    :  public Statement
    {

        virtual ~StatementListStatement() {}

        std::auto_ptr<StatementList>   statements_;

        StatementListStatement(StatementList *adoptee)
        :  statements_(adoptee)
        {
        }


        virtual StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                            FunctionScope                *functionScope,
                                            JobController                *jobController,
                                            bool                          skipArg0            // for use by the command "builtin"
                                           ) const;
            ///  Executing this function in the context of executionContext
            ///  using the specified parameters.
            ///
            ///  @return A command code that indicates what happened in the function.
            ///

        virtual std::string getDebugDescription() const 
        {
            std::string rv(" { ");

            if(statements_.get())
            {
                int count=0;

                CXXTLS_FOREACH( ThreadSafeHandle<Statement> const &cur, statements_.get()->statements() )
                {
                   ++count;

                   if(count > 1)
                       rv += " ; ";

                   rv +=  cur->getDebugDescription();


                   if(count > 5)
                   {
                      rv += " ... ";
                      break;
                   }
                }
            }

            rv += " } ";

            return rv;
        }



    }; // class StatementListStatement

    struct FunctionDefinitionStatement
    : public Statement
       ///  A statement that holds the compiled form of statements definition.  When executed,
       ///  this kind of statement does nothing.  You need to have calls to the function for
       ///  the function to be executed.
    {
        virtual ~FunctionDefinitionStatement();;
        StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                    FunctionScope               *scope,
                                    JobController               *jobController,
                                    bool                         skipArg0            // for use by the command "builtin"
                                   ) const;
        ///  @return SRC_Success unconditionally.

        virtual std::string getDebugDescription() const 
        {
           return "function ";
        }
    };

    struct FunctionScope
        /// A list of map of function name to function body mappings.  The "list" part comes from
        /// the fact that scopes are nested substructures.  Each scope sub-structure adds more
        /// names of functions but each of which is only valid in that substructure.
    {
        FunctionScope const                                   *parent_;

        typedef std::map<std::string, FunctionDefinitionStatement *>   FuncMap;
        
        FuncMap                                                members_;

        FunctionScope(FunctionScope const *parent)
        :  parent_(parent)
        {
        }

        void add(std::string const &name, FunctionDefinitionStatement *body)
           /// add a function to the scope and take ownership of the body pointer.
        {
            members_.insert(   FuncMap::value_type(name, body) );    
        }

        ~FunctionScope()
        {
            CXXTLS_FOREACH( FuncMap::value_type &f, members_)  
                delete f.second;  // delete the body
        }

        FunctionDefinitionStatement const *find(std::string const &funcName) const;
            /// Search for the specified function in this and all parent FunctionScope objects. 
            ///
            /// @return a pointer to the definition of a function or null if none is found.

    };


    struct FunctionCallStatement
    : public Statement
    {
        virtual ~FunctionCallStatement();
        std::auto_ptr< std::vector<StringExpression const *> > args_;

        FunctionCallStatement(std::vector<StringExpression const *> *args)
        :  args_(args)
        {
        }

        StatementReturnCode execute(ScriptThread::ThreadContext *executionContext,
                                    FunctionScope               *scope,
                                    JobController               *jobController,
                                    bool                         skipArg0            // for use by the command "builtin"
                                   ) const;
            /// assume args[0] is a function name and call the function that matches that
            /// name from the specified scope.  If none is found, perform a command path
            /// search and launch a child process to execute the found file in the specified
            /// execution context.
            ///
            /// The execute method will evaluate the string expressions in the args_ 
            /// vector and populate the execution context with them and then execute the
            /// statements in the function body (or call the external command).
            ///
            /// @return whatever the called function/process or return SRC_Fail if no such
            ///         function can be found.

        virtual std::string getDebugDescription() const 
        {
            std::string rv("call function ");

            if(args_.get() && !args_->empty())
            {
                std::string funcName;


                if((*args_)[0]->isSimple(funcName))
                {
                   rv += funcName;
                }
            }


            return rv;
        }
    };

    // note that many statement types can be defined and they do not need to be defined
    // in this header because they all have the same interface and they do not need to
    // call each other.  So basically an interpreter is defined by a function that knows about
    // all the statement types it can support and its parser that builds calls to the appropriate
    // statement types into a StatementList that can be executed.

};
} // namespace cxxtls
#endif
