#ifndef cxxtls_script_thread_header_included_p
#define cxxtls_script_thread_header_included_p

//
// Copyright 2002-2015, Lowell Boggs Jr.
//
// This file or directory, containing source code for a computer program,
// is Copyrighted by Lowell Boggs, Jr.  987 Regency Drive, Lewisville
// TX (USA), 75067.  You may use, copy, modify, and distribute this
// source file without charge or obligation so long as you agree to
// the following:
//
//  1.  You must indemnify Lowell Boggs against any and all financial
//      obligations caused by its use, misuse, function, or malfunction.
//      Further, you acknowledge that there is no warranty of any kind,
//      whatsoever.
//
//  2.  You agree not to attempt to patent any portion of this original
//      work -- though you may attempt to patent your own extensions to
//      it if you so choose.
//
//  3.  You keep this copyright notice with the file and all copies
//      of the file and do not change it anyway except language translation.
//
// You are responsible for enforcing your own compliance with these
// conditions and may not use this source file if you cannot agree to the
// above terms and conditions.
//
// Warning:  not all files in this directory structure may be covered by the
// same copyright.  Some of them may be part of the GNU source distribution
// and you must obey the GPL copyright for any such files.


#include <cxxtls/stringexpression.h>
#include <cxxtls/exception.h>

#include <boost/fdstream.hpp>

#include <cxxtls/file.h>
#include <cxxtls/thread.h>

#ifdef _MSC_VER
#include <windows.h>
#endif

///@file 
///  This file defines symbols that are used to implement script language
///  interpreters.  The assumption is that commands in the script language
///  are run in the context of script language thread.  This file defines
///  such thread contexts and the various functionality needed to make them
///  work.
///
///  there are two basic types of threads in a script language interpreter:
///
///  @li background jobs
///  @li pipe statements and their member stages.
///
///  Any "thread" in a script language interpreter has at least the following
///  components:
///
///   @li environment variables or a reference to them
///   @li file handles -- stdin, stdout, stderr, and any other user defined
///       variables
///   @li command line arguments to the current command
///   @li superseeded environment variables passed to the current command.
///       Usually there are only a small number of these -- not not zero.
///

namespace cxxtls
{

namespace ScriptThread
    ///  ScriptThread is the namespace that defines the top level concepts in
    ///  in a script language interpreter.  See ThreadContext.
{

    struct UserIOCancellation
       :  public Exception
       ///  An exception thrown when the user cancels I/O operation -- as on windows when the user presses ^C
       ///  while reading from a file that is not being handled as a console but actually is
    {
        UserIOCancellation()
        :  Exception("UserIOCancellation()")
        {
        }
    };

    struct PipeWriteFailure
       :  public Exception
       ///  An exception thrown when the downstream file handle in a pipe gets closed.  The exception
       ///  is thrown when the upstream code notices the file handle is closed as a pipe.  There is two
       ///  ways for this to happen:
       ///
       ///  @li  A file i/o routine notices a pipe-related write error
       ///  @li  A check of the PipeState, see below, tells you that the pipe is closed even if no
       ///       write error has occurred.
    {
        PipeWriteFailure()
        :  Exception("PipeWriteFailure()")
        {
        }
    };

    struct UnableToCreatePipe
       :  public Exception
       ///
       ///   Thrown when a pipe cannot be created.  If any handle cannot be duplicated, this error is
       ///   thrown.
       ///
    {
        UnableToCreatePipe()
        :  Exception("UnableToCreatePipe()")
        {
        }
    };



    class PipeState
       /// this object keeps track of information about a pipe between two threads.  The status
       /// information is mutex locked so that all activities can be safely performed between two
       /// threads
       ///    @li reader
       ///    @li writer
       ///
       /// For all normal uses, a ThreadSafeHandle<PipeState> should be owned by all accessors
       /// of a PipeState
       ///
       /// See comments in the terminated() member below for more information about who uses this
       /// information.
       ///
       /// Note that this class is necessary because on Windows, at least, there is no way for the
       /// write-site of a pipe to know that the pipe is closed.  So the pipe creator code, see
       /// BrashPipeStatement::execute(), must hook two threads together, using a pipe, and must also
       /// inform the FileHandleAttributes what the PipeState is for the two threads.  Then, the
       /// 
       /// Note that the FileHandleAttributes class has a function, setNoTerminatePipe(), which
       /// overrides calls to this class and other behaviors associated with pipes.  Normally,
       /// when you read from a closed file, that is a problem, but in for and while loops,
       /// the loop gets messed up and you don't want the closing of the pipe to trigger exceptions
       /// in that case.  See BourneWhileStatement::execute and BourneForStatement::execute
    {
        cxxtls::Mutex guard_;
        


#ifdef  _MSC_VER
        typedef HANDLE handle_t;
#else
        typedef int    handle_t;
#endif

        handle_t readHandle_;
        handle_t writeHandle_;
        bool     terminated_;

        bool     ignoreTerminationRequest_;   // used by the tail of of a pipe to overcome the fact that the tail end
                                              // file handles should not be terminated.  For example:
                                              //
                                              //    ( a | b ; echo stuff ) | cat
                                              //
                                              // In the above, when the pipe containing "a" and "b" finishes, b is
                                              // using the same stdout handle as the echo statement.  We do not
                                              // want the exiting of the pipe to close the stdout file handle used by
                                              // stuff
     public:

        int cancelWrite()
            /// return true if the write was successfully canceled.  If 0 is returned,
            /// call GetLastError() for an explanation
            ///
            /// you should generally not call this if teriminatePipe_ is true
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           if(writeHandle_)
           {
              terminated_ = true;
              return CancelIoEx(writeHandle_, 0);
           }

           return true;
        }

        PipeState()
        :  readHandle_(0)
        ,  writeHandle_(0)
        ,  terminated_(0)
        ,  ignoreTerminationRequest_(0)
        {
           CXXTLS_LOG("PipeState " << (void*)(this) << " Construction with writeHandle_ = 00000000, readHandle_ = 00000000" );
        }

        ~PipeState()
        {
           // try to make sure that no one is using this object and destruct it
           cxxtls::ScopedMutexLock  lock(guard_);

           CXXTLS_LOG("PipeState " << (void*)(this) << " Destruction:  writeHandle_ = " << writeHandle_ << ", readHandle_ = " << readHandle_ );

           terminated_ = true;
           readHandle_ = 0;
           writeHandle_ = 0;
        }


        bool terminated()
           /// find out if the handle has been terminated.  One side of the other must call
           /// setTerminated(true) for the state to change
           ///
           /// Up stream FileHandleAttributes objects, and the ManagedFileHandles read/write
           /// members can call this function before the start of i/o and if is true,
           /// throw a PipeWriteFailure exception to terminate the current set of script language
           /// statements.  Note any child processes using the pipes will not be using this
           /// mechanism.
           ///
           /// At this point, the following code locations are known to be interested in this
           /// information:
           ///
           ///   @li loop conditional tests (while, for, select)
           ///   @li echo statements
           ///   @li printf, when I implement it
           ///   @li sleep
           ///
           /// Whenever the stdout out file handle is terminated, if it is a pipe, these
           /// code locations should throw PipeWriteFailure, even if that doesn't make sense
           /// in order to end the loop, the writing, the wait, etc.
           ///
           /// stdin never uses this.

        {
           cxxtls::ScopedMutexLock  lock(guard_);

           return terminated_;
        }

        bool setTerminated( bool value  )
            ///  Change the terminated flag to a new value, and return the old value
            ///
            ///  Downstream pipe FileHandleAttributes objects should call this function
            ///  when the file is closed.
        {

           return false;  // temporary debugging experiment to see if this function is even necessary

           cxxtls::ScopedMutexLock  lock(guard_);

           if(ignoreTerminationRequest_)     // see ignoreTerminationRequest(), below
              return terminated_;    

           bool oldValue = terminated_;

           terminated_ = value;

           if(value)
           {
               CXXTLS_LOG("PipeState " << (void*)(this) << " TERMINATING,  writeHandle_ = " << writeHandle_ << ", readHandle_ = " << readHandle_);
           }

           return oldValue;
        }

        void ignoreTerminationRequest()
           ///  Called by pipe creation code on the very last stdout file handle of the pipe.  This call
           ///  instructs the pipe state to ignore termination requests.
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           ignoreTerminationRequest_ = true;
        }


        handle_t readHandle() 
            ///  get the read handle for the pipe
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           return readHandle_;
           
        }
        

        handle_t writeHandle() 
            ///  get the write handle for the pipe
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           return writeHandle_;
           
        }


        handle_t setWriteHandle(handle_t h)
           /// give a new value for the current write handle and return the hold one.
           ///
           /// generally is is only used to pass 0 for h
        {
           cxxtls::ScopedMutexLock  lock(guard_);


           CXXTLS_LOG("PipeState " << (void*)(this) << " setWriteHandle( old = " << writeHandle_ << ", new = " << h );

           handle_t oldValue = writeHandle_;

           writeHandle_ = h;

           return oldValue;
          
        }

        handle_t setReadHandle(handle_t h)
           /// give a new value for the current read handle and return the hold one.
           ///
           /// generally is is only used to pass 0 for h
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           CXXTLS_LOG("PipeState " << (void*)(this) << " setReadHandle( old = " << readHandle_ << ", new = " << h );

           handle_t oldValue = readHandle_;

           readHandle_ = h;

           return oldValue;
          
        }

        bool hasReader(handle_t h)
           // return true if the write handle in this pipe state object equals h and
           // if the read Handle is null
        {
           cxxtls::ScopedMutexLock  lock(guard_);

           if(h != writeHandle_)
              return true;  // technically, we don't know

           // at this point, h == writeHandle_

           if(readHandle_ != 0)
              return true;  // there is a reader

           return false;  // this is the correct pipe state object but we have no reader

        }


    };  // PipeState


    
    struct ManagedFileHandles
       /// This object controls the file handles that are associated with a script
       /// thread.  On Windows, HANDLE objects are kept as file handles, but on
       /// unix integers are used.  
       ///
       /// The standard input, standard output, and standard error file handles are
       /// included in this list ANY APPLICATION USING THIS CLASS MUST NOT WRITE
       /// DIRECTLY TO THE STANDARD HANDLES EVERY!  Instead, you should either write
       /// to log files using the ManagedFileHandles object itself, or you should
       /// use dup() or DuplicateFile() to create your own copies of these file handles
       /// and write to those copies.  However, in a multithreaded environment, you
       /// might as well use ManagedFileHandles to do the thread safety for you.
       ///
       /// Warning:  this function also controlls the current directory.  If you are
       /// going to use some other thread or functionality that changes the current
       /// directory, that functionality must interact with ManagedFileHandles to set
       /// the current directory for you.
       ///
       /// Also note that any environment variable settings you make outside the scope
       /// of the ScriptExpr::ExecutionVariables class will not be passed to child
       /// processes -- or to any functionality making use thereof.
       ///
       /// Note that "script file handles" are not the same as their operating system
       /// counter parts.  The script language keeps track of integers as file handles
       /// but these integers do not map to the same value in the underling operating
       /// system file handles.  For example, script file handle 0 may map to 37 on unix, or it
       /// map to 0x10347 on Windows.
       /// 
       /// In a given script, script file handles 0-2 are should be treated by the script
       /// writer as STDIN, STDOUT, and STDERR, but these "script handles", will only actually 
       /// be mapped to the corresponding process values during a very brief window in which the 
       /// child processes are created.
       ///
       /// Other file handles can be used with more abandon.  
       ///
       /// Note Windows and unix/linux
       /// deal with the standard file handles in different ways but there is a common
       /// theme in the way they work:  In both cases:
       ///
       ///    @li there is one set of standard handles per process
       ///    @li there is a way to force a file handle to be used as one of that set
       ///
       /// On Windows, any file handle can be asserted to STDIN, STDOUT, or STDERR using the
       /// function, SetStandardHandle(handle, ID), where ID is one of:
       ///
       ///    @li STANDARD_INPUT_HANDLE
       ///    @li STANDARD_OUTPUT_HANDLE
       ///    @li STANDARD_ERROR_HANDLE
       ///
       /// However
       /// on unix/linux, only the file handle 0, 1, and 2 can occupy the standard handles role.  
       /// But if you assume that the script file handles are never mapped to the unix/linux
       /// os file handles, 
       /// then you can force any file handle to be a standard handle by using the dup2() 
       /// function:  All you have to do is write code like this:
       ///
       ///@begincode
       ///     dup2(existingfileHandle, desiredStandardHandle)
       ///@endcode
       ///
       /// Where desiredStandardHandle is either 0, 1, or 2.  You should check the return code
       /// from dup2() to confirm that you got 0, 1, or 2 as desired.  The above code will close
       /// the existing standard handle and create a new cloned version existingFileHandle with
       /// the desiredStandardHandle value.
       ///
       /// Note that the standard application file handles should be viewed as a precious resource
       /// and should not in general be used explicitly the script interpreter for anything but
       /// the above purpose.  
       ///
       /// The ManagedFileHandles class keeps track of the system standard handles resources and
       /// and uses mutexes to guard against their mistaken use.
       ///
       /// A given thread may or may not own/manage its ManagedFileHandles object -- and the
       /// caller must keep track of that outside this object.  You cannot copy or
       /// overwrite ManagedFileHandles objects.  This is done for speed reasons.  If a
       /// command does not change stdout, stderr, or stdin, we don't want to have to 
       /// pay the penalty of duplicating the file handles.  But if any of these things
       /// are changed, the duplication must occur.  That is accomplished by creating a
       /// new ManagedFileHandles object and duping() all the old values and creating new
       /// ones which can be closed when the duplication goes out of scope.  For example,
       /// in bash if you do one of the following:
       ///
       ///@begincode
       ///
       ///    commandName >outputFile
       ///    commandName >outputFile <inputFile
       ///    commandName >&1 </dev/null
       ///
       ///@endcode
       ///
       /// At the time commandName executes, there is a temporary superseed of stdout.  after
       /// the command terminates, the superseed goes out of scope.  So, the execution of
       /// commandName must clone its ManagedFileHandles, changing the stdout file handle in
       /// the process.  After commandName finishes executing, the code must destroy the modified
       /// clone and go back to using the origninal stdandard handles.
       ///
       /// Use the nested typedef iterator, and the nested functions, begin and end, to iterate
       /// over the file handles.
    {


        private:

            ManagedFileHandles(std::string const &curdir);                // use create() rather than construction
            ManagedFileHandles(ManagedFileHandles const &rhs);            // you can't copy these things
            ManagedFileHandles &operator=(ManagedFileHandles const &rhs); // no can assign neither dog

            ~ManagedFileHandles() ;

                ///  close the file handles and free all resources associated with this
                ///  object.  Called only by the destroy() function

        public:

        static ManagedFileHandles   * mainStdHandles;       ///<  on entry of the program, these are the standard file handles.

        static Mutex currentDirectoryMutex_;                ///<  a guard variable for protecting temporary changes to the
                                                            ///   current directory.  Use the class ScopedCurrentDirectoryUsage
                                                            ///   to temporarily hold off other threads from changing the current
                                                            ///   directory.  

       struct ScopedCurrentDirectoryLock  
           /// declare a variable of this type in order to temporarily lock out other threads from changing the
           /// current directory -- and thus causing thread corruptin in the chdir() function.
           ///
           ///  Declare variables of this type only limit tiny sections of code that open a file or launch a background job
           ///  or whatever.  Note that the change is super-temporary.  Don't expect your current directory not to be
           ///  jerked out from under you.  Always keep your desired current directory in a variable and use this helper
           ///  class to let you set it.
           ///
           ///  usage example:
           ///
           ///@begincode
           ///
           ///   int fh;
           ///
           ///   {
           ///       ManagedFileHandles::ScopedCurrentDirectoryLock  lock;
           ///
           ///       chdir(myExpectedCurrentDirectory);
           ///
           ///       fh = open("file", O_RDONLY);
           ///
           ///       // STOP:  do not read the file here!  do that outside the curly braces!
           ///   }
           ///
           ///
           ///@endcode
           ///
           ///   fh is now opened in the context of your expected current directory, but who knows what other threads might
           ///   have done to the actual current directory.

       {
           ScopedMutexLock lock_;

           ScopedCurrentDirectoryLock()
           :  lock_(ManagedFileHandles::currentDirectoryMutex_)
           {
           }
       };



        struct FileHandleAttributes
            /// A structure that defines and maintains inheritable file handles that are passed to a child processed.
            /// This class is just a wrapper around operating system identifiers for file handles that are passed.
            /// File handles are identified by a number, 0 to inifinity.  Numbers 0 to 2 are special cases that are
            /// mapped to stdin, stdout, and stderr just as is found in unix.  In windows, these same handles are
            /// mapped to STANDARD_INPUT_HANDLE, STANDARD_OUTPUT_HANDLE, and STANDARD_ERROR_HANDLE.
            ///
            /// Note:  the destructor does not close the file handles!  The owner must do that using closeIfNotError().
            ///
            /// This class exists as a kind of wrapper around functionality and is not a full fledged value class.
        {
           long errno_;  // most recent error number
           int fh_;  // the number used by scripts to access this particular file handle for reading and writing.

           bool error_;  // just a flag indicating that a file-open problem have occured.  If a reopenAs() operation
                         // fails, this won't get set because the old file handle will remain in effect.

#ifdef _MSC_VER
           typedef HANDLE  handle_t;

           HANDLE msHandle_;  ///< corresponding microsoft handle for file fh used in scripts.

           handle_t getOSHandle() const { return msHandle_; }

           bool  isPipe_; ///< true if the this is a DOWNSTREAM pipe.  upstream pipes are not knowable

#else
           typedef int  handle_t;
           int    unixHandle_;  ///<  corresponding unix file handle for this "logical" file handle.

           handle_t getOSHandle() const { return unixHandle_; }

#endif
           bool input_;       ///< true if this an input file handle that a child process can read from
                              ///  false means that the child can only write to the file handle.


           // the following are for readLine()
           
           char buffer_[8129];
           size_t bytesInBuffer_;

           cxxtls::ThreadSafeHandle<PipeState>  pipeState_;  // defaults to null

           bool terminatePipe_;  // When the downstream handle is closed, we normally cancel 
                                 // writes to the upstream
                                 // pipe.  Normally, you want to do this, but in the "do" block of a 
                                 // while statement and for statement, you don't.  The while statement 
                                 // might be dependent
                                 // on the reading of stdin.  In that case, automaticaly terminating
                                 // the pipe (as would happen when you exit a command with redirections
                                 // will stop the loop by mistake.

           void setNoTerminatePipe(bool value = true)
               ///  called by BourneWhileStatement::execute and BourneForStatement::execute to suppress
               ///  normal pipe closure at end of file in while and for loops.
               ///
               ///  value == true means turn it off...
               ///
               ///  Sorry, the code was already named no terminate pipe before I realized that I needed
               ///  set it both ways.  idiot...
           {
              terminatePipe_ = !value;  
           }

           void setPipeState(cxxtls::ThreadSafeHandle<PipeState> &s)
              /// point the pipeState_ to the correct object.  This is only used
              /// by the code that creates pipes in the higher level languages.
              /// Normally, the pipeState_ is null.  See BrashPipeStatement::execute()'s use
              /// of this function.
           {
               pipeState_ = s;
           }

           FileHandleAttributes()
           :  errno_(0)
           ,  fh_(-1)
           ,  error_(true)
#ifdef _MSC_VER
           ,  msHandle_(0)
           ,  isPipe_(false)
#else
           ,  unixHandle_(-1)
#endif
           ,  input_(0)
           ,  bytesInBuffer_(0)
           , terminatePipe_(true) // normally you want to terminate the upstream pipe when the downstream pipe
                                  // closes, but when you are in a while loop, the stdin pipe should not be so closed.
           {
           }



           ~FileHandleAttributes()
           {
               // do not close the file automatically -- we pass these objects around and the file needs to be open when we do.
               error_ = true;
           }

           FileHandleAttributes clone();  
                                          ///< use the operating system file handle duplication logic
                                          /// to create a new FileHandleAttributes object with different
                                          /// fh_ values refering to the same physical file, but with a
                                          /// different operating system file handle.
                                          ///
                                          /// Note that the caller is responsible for calling closeIfNotError
                                          /// when we are done with the copy.  But the destructor does not do it.

           void closeIfNotError();    ///< close this file handle if not already closed and if not an error.


           bool reopenAs(std::string const &filename, bool atEnd=false); 
               ///< close and reopen this file handle with new name.  Return true if successful.  Otherwise,
               /// the prev handle remains in effect .  
               ///
               /// Note that if the file is currently one of the following:  stdin, stdout, or stderr, this function
               /// gaurantees that the newly opened file will also play the corresponding role.  Which means that this
               /// function changes the application wide stdin, stdout, or stderr.
               ///
               /// Consequently, caution must be used when dealing with multiple threads.  The ownership of the system wide
               /// resources be controlled via mutex lock.  When executing bash script language commands like the following:
               ///
               ///@begincode
               ///
               ///     programName parms >output < input 2> error
               ///
               ///@endcode
               ///
               ///  The thread should temporarily take control of the stdin, stdout, and standard error handles, make the call
               ///  to launch the child process, then restore the previous values.



            bool reopenDup(handle_t os_fh);
               ///<  close and reopen this file handle as a duplicate of an existing operating system file handle.
               ///   See getOSHandle() methods above for ways to get the file handle for the other file

            bool good() const
               /// return true if this file handle is working properly.
            {
                return (fh_ != -1) && !error_;
            }


            bool readLine(std::string &line); 
                /// read up to 1 line of text into the output line.  
                ///
                /// @return true if successful.
                ///
                /// @b Throws:  UserIOCancellation when the user presses ^C 


            bool isConsole() const;
                /// return true only if this file handle represents a windows console.


        };  // FileHandleAttributes

        void setNoTerminatePipe(int fh, bool value=true);   ///< call noTerminatePipe on the file attributes of handle fh.
                                                            ///  true means to turn off pipe termination.  false turns it on.

        FileHandleAttributes::handle_t getOSHandle(int fh); ///< get the raw os handle for a given "file handle"

        void setPipeState(int fh, cxxtls::ThreadSafeHandle<PipeState> &s);
            ///<  Change the pipe state pointer in the specified file handle.
            ///   This should only be called by higher level language implementation routines that set up pipes

        void terminatePipe(int fh); ///< mark the pipe associated with a file handle as terminated


        static FileHandleAttributes::handle_t nulHandle_;   ///<  handle to the nul: device (/dev/null)

        static FileHandleAttributes getStdinFileHandleAttributes();  /// On unix, just return fh_ = 0 and input_ = true, but on windows, 
                                                                     /// return  include msHandle_ 

        static FileHandleAttributes getStdoutFileHandleAttributes();  /// On unix, just return fh_ = 1 and input_ = false, but on windows, 
                                                                      /// return  include msHandle_ 

        static FileHandleAttributes getStderrFileHandleAttributes();  /// On unix, just return fh_ = 2 and input_ = false, but on windows, 
                                                                      /// return  include msHandle_ 

        ManagedFileHandles *clone();
           ///< Return a clone of the ManagedFileHandles object.
           ///
           /// A clone has different FileHandleAttribute objects but they refer to the same
           /// physical streams of data.  A copy would have identical numbers.
           ///
           /// The caller owns the return pointer and must arrange for its deletion at the proper time.

        ManagedFileHandles *pipe();
           ///<  Similar to clone but this function creates a communication relationship between
           ///   the returned ManagedFileHandles object and *this.  The relationship works like
           ///   this:  anything written to the standard output handle of the the returned object
           ///   will be read from the standard input handle of *this.
           ///
           ///   You should probably only call this function on a freshly cloned object because
           ///   it affects *this in a way that may not be desirable to other script stack frames
           ///   using the same *this object.

        static void destroy(ManagedFileHandles *victim)
            /// normally, you don't own your ManagedFileHandles object but sometimes you do and
            /// need to destroy them.
           {
               delete victim;
           }

        static ManagedFileHandles *create(std::string const &curdir) { return new ManagedFileHandles(curdir); }
            /// create an empty ManagedFileHandles object.  Call addFileHandles() to populate your
            /// new object.

        int callProcess(std::string const &commandLine, StringExpr::ExecutionVariables const &env, bool isPipeMiddle);
            ///
            /// threadsafe
            ///
            /// Invoke a command line with a particular environment variable set and passes the
            /// standard file handles to the process. 
            ///
            /// a call is made automatically to grabStandardHandles.  The old value of the handles are
            /// not saved.
            ///
            /// On windows, the command line will be prefixed, automatically, with "CMD /C " and the
            /// environment variables will be converted into a set of calls to setenv() before
            /// the invocation.
            ///
            /// Note that on Windows, you must used CMD /C, if you want the child process to properly
            /// inherit the STD file handles.
            ///
            /// @return the exit code from the process or a negative number indicating error.
            ///
            ///   @param[in] commandLine A character string to be parsed into parameters
            ///   @param[in] env         The environment variables to pass to the child process
            ///   @param[in] isPipeMiddle A flag indicating that this invocation refers to the middle of
            ///                           a pipe chain - or to its end, but not to its beginning.

        void addFileHandle(FileHandleAttributes const &h)
            ///  Call this using some function that returns a proper object.  Typically, getStdinFileHandleAttributes() - 
            ///  getStderrFileHandleAttributes().  But other functions can return this.  Note that the file handles so
            ///  created should be inheritable.
        {
            handles_.insert(std::pair<int,FileHandleAttributes>(h.fh_, h));  
        }

        enum IO_ERROR
        {
           IOERR_OK=0,
           IOERR_bad_file,
           IOERR_closed_file,
           IOERR_not_opened_for_read,
           IOERR_other_os_error,
           IOERR_not_opened_for_write

           // note: pipe write failure needs to be handle as an exception so that it doesn't get lost
           // but often we are going to silently ignore it
        };

        static char const * errorString(IO_ERROR); ///<  @return an error message string given an error number.

        bool readLine(int fh, std::string &line); ///<  Read up to a newline or end of file.


        IO_ERROR readByte(int fh, char *bytePtr)                            ///< @return IOERR_OK if successfull and return the character through *bytePtr
        {
            size_t count=0;

            IO_ERROR rv = readBytes(fh, 1, bytePtr, &count);

            if(rv)
               return rv;

            if(count == 0)
               return IOERR_other_os_error;

            return IOERR_OK;
        }


        IO_ERROR readBytes(int fh, size_t maxBytes, char *bytePtr, size_t *bytesRead);
                                                                            ///< @return IOERR_OK if successfull and 
                                                                            /// return at most maxBytes the characters through *bytePtr.  Does not stop
                                                                            /// at newline if this is a disk file.
                                                                            ///
                                                                            /// returns number of bytes thread through *bytesRead


        IO_ERROR writeBytes(int fh, std::string const &buffer)   ///< @return IOERROR_OK if successful.   
                                                                 /// write a line of text followed but do not add newline
                                                                 ///
                                                                 /// fh is not a system file handle but
                                                                 /// rather a key that is used to find the FileHandleAttributes object in *this.
        {
            return writeBytes(fh, buffer.data(), buffer.size());
        }

        IO_ERROR writeBytes(int fh, char const *buffer, size_t length);  
                                                                 ///< @return IOERROR_OK if successful.   
                                                                 /// write a line of text followed but do not add newline
                                                                 ///
                                                                 /// fh is not a system file handle but
                                                                 /// rather a key that is used to find the FileHandleAttributes object in *this.

        IO_ERROR writeByte(int fh, char byte)                    ///< @return IOERROR_OK if successful.
                                                                 ///  write 1 byte to the specified file handle.  
                                                                 ///
                                                                 /// fh is not a system file handle but
                                                                 /// rather a key that is used to find the FileHandleAttributes object in *this.
        {
           return writeBytes(fh, &byte, 1);
        }
                        
        bool reopenAs(int fh, std::string const &filename, bool atEnd=false);   ///<  re-opens the specified file handle using the new
                                                                                ///   file name.  If this is an input file, the filename
                                                                                ///   must be readable.  If it is an output file handle
                                                                                ///   the file name must be writable.  atEnd mean seek to the
                                                                                ///   end before writing, if this is an output file handle.


        bool reopenDup(int modified_fh, int copyfrom_fh);   ///<  re-opens the current file hande as a duplicate of some other file handle.


        typedef std::map<int, FileHandleAttributes>::iterator       iterator;
        typedef std::map<int, FileHandleAttributes>::const_iterator const_iterator;

        iterator begin() { return handles_.begin(); }
        iterator end()   { return handles_.end();   }

        const_iterator begin() const { return handles_.begin(); }
        const_iterator end()   const { return handles_.end();   }

        iterator       find(int fh)       { return handles_.find(fh); }
        const_iterator find(int fh) const { return handles_.find(fh); }


        std::string const &currentDirectory() const 
        { 
            return currentDirectory_;  // keep this as a separate source line so that the debugger can print it.  if the whole thing
                                       // lies on one line, the debuger often cannot print the string because we are not yet
                                       // sitting on the line of code inside the function body, and this is not initialized.
        }

        std::string mapFilename(std::string const &file) const;

        bool               setCurrentDirectory(std::string const &path_in)
            /// not thread safe!
            ///
            /// Try to change to the specified directory.  If that fails, return false.
            /// If successfull, record the path for this thread and return true.
        {

            std::string path(path_in);

            ScopedCurrentDirectoryLock lock;


#ifdef _MSC_VER
            SetCurrentDirectory(currentDirectory_.c_str());  
                                               // First set the current directory to the
                                               // place we think it should be.  We need to
                                               // do that to account for the fact that 
                                               // other threads WILL have changed it since
                                               // this thread set it last.  
                                               //
                                               // Next, we change directory again in the
                                               // context of the existing current directory
                                               // to allow relative paths to work correctly.

            FileName::sorted_names_t  matches; // matching file names

            if (path.find_first_not_of("\\/.") >= path.size())
            {
               // path contains only dots and slashes

               FileName tmp(path);

               tmp.convert_to_absolute_path();

                path = tmp;
            }

            // path is now a full path.
                     
            if (!path.empty() && (path[path.size()-1] == '/' || path[path.size()-1] == '\\'))
            {
                // if the path ends /, then assume it is a directory and just stick in matches

                matches.insert(path);
            }
            else
            {
                // the path does not end in /, so lets see if there really is a directory by the specified name

                FileName::find_matching(path.c_str(), &matches, 0x11); // include full pathanmes in the result, and do case insensitive
            }

            // if there is no directory by the specified name return failure


            if(matches.size() != 1)
            {
                return false;

            }

            // now try to cd to the specified directory


            if(0 != SetCurrentDirectory(matches.begin()->c_str()))
            {
                char buffer[2000];

                GetCurrentDirectory(sizeof(buffer), buffer);

                currentDirectory_ = buffer;

                return true;
            }
#else
            chdir(currentDirectory_.c_str());  // First set the current directory to the
                                               // place we think it should be.  We need to
                                               // do that to account for the fact that 
                                               // other threads WILL have changed it since
                                               // this thread set it last.  
                                               //
                                               // Next, we change directory again in the
                                               // context of the existing current directory
                                               // to allow relative paths to work correctly.
            if(0 == chdir(path.c_str()))
            {
                char buffer[2000];
                ::getcwd(buffer, sizeof(buffer));   // finally, we get the full pathname of
                                                    // the new current directory for later
                                                    // queries by the application using this
                                                    // class.
                currentDirectory_ = buffer;
                return true;
            }
#endif

            // we could not cd to the specified directory, so return false;


            return false;
        }

        size_t numHandles() const { return handles_.size(); }  ///<  Get the count of file handles


        FileHandleAttributes getFileAttributes(int fh) const
            /// get the attributes of the specified file handle.  Call the good() member function to see
            /// if this is valid, working file handle.
        {

            const_iterator where = handles_.find(fh);

            if(where == handles_.end())
               return FileHandleAttributes();   // an error return code.

            return where->second;  // the actual value
           
        };


        private:


            // STOP:  if you add new data members here, ask your self how the clone() method will deal
            //        with them.  handles_ for example is not handle via simple copy, but the currentDirectory_
            //        is

            typedef std::map<int, FileHandleAttributes>  HandlesVector;

            HandlesVector handles_;

        
            std::string currentDirectory_;   

            // see STOP comment above.


    };  // class ManagedFileHandles


    struct ThreadContext
        ///  A data structure that contains the top level resources for a 
        ///  a script thread's execution.  A "script thread" is the structure that
        ///  gets executed by a thread when the script language interpreter creates
        ///  a high level "thread of execution".  In Bash, there are no "threads"
        ///  but there are "jobs" which act like threads.  A ScriptThread is a real
        ///  thread that acts like a job -- kind of.  The goal here is to be able 
        ///  to define separate streams of execution in a portable way that will allow
        ///  user defined scripts to run the same regardless of environment and get
        ///  good performance, though not necessarily maximal performance, in all
        ///  environments.  
        ///
        ///  A script thread has the following major components:
        ///
        ///  @li variables
        ///  @li functions
        ///  @li file handles (and the ability to launch child processes which use them!)
        ///
        ///  In bash, a "job" is a separate stream of execution which is created when the
        ///  user issues an ampersand command -- that is a command whose invocation
        ///  ends in "&".  
        ///
        ///  Each script job has a "thread of execution" which is defined in terms of
        ///  a "script runtime language interpreter" which is not defined by the
        ///  the Scriptthread.  Functions are implemented in terms calls to C++
        ///  language virtual functions which implement the "script language intepreter",
        ///  whatever that might be.
        ///
        ///  Script Language File handles are logically "numbers" which represent low level streams of 
        ///  i/o operations.  Both on Windows and on Unix/linux, file handles in a parent
        ///  process may either be private or they may be passed to a child process
        ///  at the time of its execution.  This class, ThreadContext, keeps track of the
        ///  the file handles that are defined to be passed to the child.  It also provides
        ///  buffered i/o interfaces to them, as needed.  File handles 0, 1, and 2 on Unix
        ///  linux are mapped to the standard input, standard output, and standard error
        ///  functionality.  On Windows, there are corresponding functions that may be called
        ///  to assign any file handle one of those roles.  However, the ThreadContext will
        ///  maintain the unix standard meaning for those file handles.  The windows
        ///  functions involved in this will be:
        ///
        ///  @li SetStandardHandle(type, fh)
        ///  @li DuplicateHandle()
        ///  @li CreateFileHandle()
        ///
        ///  On Unix/Linux the following will be involved:
        ///
        ///  @li open()
        ///  @li dup()
        ///  @li dup2()  
        ///
        ///  Note that in both environments, the act of temporarily cloning the process'
        ///  "standard handles" requires that threads be locked to prevent confusion.  That
        ///  is, whenever a new thread is being created, all access to the standard files 
        ///  is shut off.  This requires that the C++ functions that implement the thread
        ///  i/o operations to the standard file handles must use mutex locks at all times.
        ///  Thus using file handle 0, 1, and 2 directly or even indirectly are strictly
        ///  forbidden in all environments.  This means that low level runtime functions
        ///  that perform unwrapped i/o should be treated with extreme caution.
        ///
        ///   The function "system" is verbotten as they say.  Nor should you write code that
        ///   uses CreateProcess(), fork(), exec() or whatever.  The only way to run child
        ///   processes is through the ManagedFileHandles class which has a method called
        ///   "callProcess()".  That is "call" not "start".  A thread which launches a child
        ///   process must wait for it to terminate.    A script language can hide this fact
        ///   from the user though.  Consider how bash does this:  if you want background
        ///   operations, you use use "&" operator on a statement.  Bash launches that statement
        ///   as a background job.  Bash keeps track of it so you don't have to.
        ///   Bash uses asynchronous process commands like  "wait" to see what is going on with
        ///   background jobs.  However, in this implementation, the background job is actually
        ///   a running thread that waiting for the child.  When the child terminates, the thread
        ///   continues on to the next thing it is supposed to be doing.  The jobs functionality
        ///   in bash can be simulated with functions that interact with the running threads.
        ///  
        ///   The current directory:
        ///
        ///   In the original bash implementation, the current directory is specific to the process id
        ///   and since a fork/exec operation was used to handle background jobs, it was trivial to
        ///   deal with the current directory.  However, this is not the case with threads -- there
        ///   is no "thread current directory", there is only the application's current directory.
        ///
        ///   The current directory is only interesting when you are opening files or launching child
        ///   processes.  The class, ManagedFileHandles, is where you register the current thread's
        ///   current directory.  Whenever you open a file in the script language, and whenever you
        ///   launch a child job, the "current directory", must be temporarily superseeded with the
        ///   current threads value, thereof, and this must be done in a thread safe manner.
        ///
        ///   This is only possible if no code in any thread directly acts on the current directory.
        ///   on its own.  No code in your script language should use chdir() or SetCurrentDirectory
        ///   or whatever.  Even calling getcwd() is dangerous -- you should get that information
        ///   from the ManagedFileHandles object in the thread context. 
        ///   
    {
        std::auto_ptr<StringExpr::ExecutionVariables> clonedEnvironment_;   // usually null
        StringExpr::ExecutionVariables* env_;         // used by not owned.
        StringExpr::ExecutionContext   varContext_;
        ManagedFileHandles            *fileHandles_;  // pointer to the object that can read/write files
                                                      // and execute programs.  It also keeps track of the
                                                      // current directory.
        bool                           execLog_;      // are we currently logging execution?

    private:

        // You cannot construct these directly, use create() and destroy()

        // the following is meant for cloning in the current job for file handle redirections

        ThreadContext(std::string const &                               scriptName,
                      StringExpr::ExecutionContext::ArgsVector const &  argsVector,
                      StringExpr::ExecutionVariables                   *startEnv,   /* use but do not take ownership */
                      ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
                     );
            ///< PRIVATE:  Construct an atomic ThreadContext given
            /// @param[in]       scriptName The name of the script being executed in this thread
            /// @param[in]       args      The command line argument vector, possibly null
            /// @param[in,out]   startEnv  A pointer to the environment, someone else owns, but we use.
            ///

        // the following is meant for cloning for other jobs
        ThreadContext(
                      StringExpr::ExecutionVariables                 &  parentEnv,   /* copied into clonedEnvironment_ */
                      std::string const &                               scriptName,
                      StringExpr::ExecutionContext::ArgsVector const &  argsVector,
                      ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
                     );
            ///< PRIVATE:  Construct an atomic ThreadContext given
            /// @param[in]       scriptName The name of the script being executed in this thread
            /// @param[in]       args      The command line argument vector, possibly null
            /// @param[in,out]   startEnv  A pointer to the environment, someone else owns, but we use.
            ///

        ~ThreadContext()
            /// PRIVATE:  destruct a given ThreadContext
        {

           CXXTLS_DEBUG("~ThreadContext(" << (void*)(this) << ")" );

           ManagedFileHandles::destroy(fileHandles_);  // obviously fileHandles_ passed to the constructor
                                                       // must be cloned from something else for this
                                                       // to work
        }

        ThreadContext(ThreadContext const &);            // not implemented,
        ThreadContext &operator=(ThreadContext const &); // not implemented

        void operator delete(void *p)
        {
            ::operator delete(p);
        }

    public:
        
        static
        ThreadContext* 
        create(std::string const &                               scriptName,
               StringExpr::ExecutionContext::ArgsVector const &  argsVector,
               StringExpr::ExecutionVariables                   *startEnv,    /* use but do not own */
               ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
              )
        ///  Use member create() to create ThreadContext objects and destroy() to discard them.
        ///  This function is mean to create a thread context for the same "job" that you are already in.
        ///  Think of this as being in the same thread, which it is.
        {
            return new ThreadContext(scriptName, argsVector, startEnv, fileHandles);  // creates a thread safe copy (atomic copy)
        }

        static
        ThreadContext* 
        createForNewJob
               (std::string const &                               scriptName,
               StringExpr::ExecutionContext::ArgsVector const &  argsVector,
               StringExpr::ExecutionVariables                   *startEnv,    /* use but do not own */
               ManagedFileHandles                               *fileHandles /* ownership taken by this object*/ 
              )
        ///  Use member create() to create ThreadContext objects and destroy() to discard them.
        ///  Think of this as "clone for pipe".  It is meant for use in creating thread contexts for other threads.
        {
            return new ThreadContext(*startEnv, scriptName, argsVector, fileHandles);  // creates a thread safe copy (atomic copy)
        }


        static
        void 
        destroy(ThreadContext* victim)
            /// use destroy() to delete ThreadContext objects and create() to create them.
        {
            delete victim;
        }

        void setExecutionLogging(bool on) { varContext_.setExecLog(on); }
        bool loggingExecution() const { return varContext_.execLog(); }

        ThreadContext* cloneForSameJob
                            (
                              std::string const &scriptName,
                              StringExpr::ExecutionContext::ArgsVector const &  argsVector
                            ) 
                            ;        ///< copy this ThreadContext into a new
                                     ///  object with the following properties:
                                     ///  @li  it shares
                                     ///       itself therefrom.
                                     ///  @li  it clones the ManagedFileHandles
                                     ///       into a new item pointing at the
                                     ///       physical files with different
                                     ///       file handles.
                                     ///  
                                     ///  Superseeding the script name and argsVector in the
                                     ///  process
                                     ///
                                     ///  It does not copy the environment, but uses the
                                     ///  same environment as the parent

        ThreadContext* cloneForNewJob();
                                     ///< copy this ThreadContext into a new
                                     ///  object with the following properties:
                                     ///  @li  a new environment is created and owned by the
                                     ///       clone.  It is identical to the parent but 
                                     ///       is separate.
                                     ///  @li  it clones the ManagedFileHandles
                                     ///       into a new item pointing at the
                                     ///       physical files with different
                                     ///       file handles.
                                     ///  
        ThreadContext* pipe()
                           ;   ///<  Similar to clone but creates a relationship between 
                               ///   the returned ThreadContext and *this.  The relationship
                               ///   works like this:  The standard out file handle from the
                               ///   returned pointer is piped to the standard input file
                               ///   handle of *this.   
                               ///
                               ///   The purpose of this function is to create a communications
                               ///   relationship between two threads that might be writing
                               ///   to standard out in one thread but a different thread can
                               ///   read and use the data so written.
                               ///
                               ///   a "pipe" acts like a double ended queue which can be
                               ///   accessed via file handles.  The assumption is that data
                               ///   will only flow in one direction.
                             
         StringExpr::ExecutionContext &variables() { return varContext_; }  ///<  get a writeable references to the threads variables

         std::pair<std::string,int> findInExecHash(std::string const &name, int hashEntityType)          
             ///   used to find programs in the PATH list
         {
            return variables().variables().findInExecHash(name, hashEntityType);
         }

         void addToExecHash(std::string const &commandName, std::string const &filePath, int hashEntityType)
             ///   used to add brash scripts, with the .brash extension, to the hash.
         {
              variables().variables().addToExecHash(commandName, filePath, hashEntityType);
         }

         void        clearExecHash()
         {
            variables().variables().clearExecHash();
         }

         std::map<std::string, std::pair<std::string,int> > &execHash() 
             ///  get the current program name hash table
         {
             return variables().variables().execHash();
         }


    };


    class Job
        ///  A script language interpreter job that can be performed in a thread.  
        ///  All scripts have
        ///  at least one job, the main job for the script.  That job can spawn other
        ///  jobs.
        ///
        ///  This is a base class from which you should derive.
        ///
        ///  You should superseed member run() and put the code in there that will execute
        ///  the job in a child thread.  
        ///
        ///  To start the thread running, invoke the start() member and keep track of the
        ///  returned ThreadId because you'll need to call joinThread() on it.
    {
    public:
        ThreadId start() 
        {
           // start the thread running.  The thread will invoke run() and when it finishes
           // it will delete this.

           return createThread(&Job::start, this);
        }


    protected:


         ThreadContext *context_;  // variables, file handles, etc.

         Job(ThreadContext *p)
         :  context_(p)
             ///  Create the job with a given context and let the job delete the context
             ///  when it has finished executing.  
         {
         }

         static CXXTLS_DECLARE_THREAD_FUNCTION(start,jobPtr);  // this function executes the
                                                             // the functionality of the 
                                                             // thread.  You must pass
                                                             // a "Job" object to it and
                                                             // the job pointer will be
                                                             // deleted when the job is done.

         virtual void *run()= 0; ///<  The virtual method that will be executed to perform
                                 ///   your specific type of job.  Return a pointer to tell
                                 ///   someone somewhere what happend.  The return value will
                                 ///   likely be ignored so returning 0 is good -- unless you
                                 ///   specifically know better.
                                 ///
                                 ///   This function has no parameters.  To pass it parameters
                                 ///   you should populate your derived-class-Job-object with
                                 ///   data members that you know how to use as parameters.

         virtual ~Job()
         {
             // automatically invoked when run() completes.

             ThreadContext::destroy(context_);
             context_ = 0;
         }
    };


}  // ScriptThread
}  // namespace cxxtls
#endif
